# -*- coding: utf8 -*-
'''

FRaC library functions

By convension, the 'type' of nominal features is a list of strings (all
possible values) and the type of continuous features  is a tuple (minimum,
maximum value).  Thus, given feature type, programs can test if
type(feature_type)==list or type(feature_type)==tuple to determine feature
type.

By convension, missing values are indicated by a NoneType value.

Copyright © 2010 Keith Noto.  Disributed under Gnu General Public License (see http://www.gnu.org/licenses/gpl.html).

'''
__author__ = "Keith Noto <noto@cs.tufts.edu>";
__license__ = "GPL";
__copyright__ = "Copyright © 2010 Keith Noto";
VERSION = __revision__ = "1.2";
	# 1.0	2010/Apr/01	Original, with lsf module and grid search code (commented out template)
	# 1.1	2010/Jul/08	Modified to compute NS one feature at a time, then clean
	# 1.2	2010/Oct/01	Modified slightly to return a list of NS, let the caller do the summation

import sys, math, time, copy
from util import *;

verbose = open(os.devnull, 'w') # module-wide output for warnings/etc.
verbose = sys.stderr

# Generate the normalized surprisal anomaly scores for a set of query points.
#
# X: Given normal examples--M training vectors, each of length N
# Q: Test set
# t: Types of each of N features.  Types include possible values
# F:  set of features for which to learn predictors
# rho:  rho[i] is a function reducing x_j to the vector of [potential] predictors of target feature i
# model:  A function that maps a list of (x,y) pairs to a function f(x) -> predicted y 
#
# return:  scores; pairs of (average match count, average probability) for each query example (CFA)
def match_scores(Xorig, Qorig, torig, F, rho, model):

	t0 = time.time()
	progress = lambda msg : verbose.write("# @%.2f seconds:  %s ...\n" % (time.time()-t0, msg))

	L = 0 # count number of  valid features (invalid features have type None, indicating they're to be ignore, all const or something)

	matchcount = [0]*len(Qorig)		# number of matching feature predictions per testset example
	total_prob = [0.0]*len(Qorig)	# total probability of observed feature values given predictions (per testset example)

	for i in range(len(torig)): # for each feature

		if i not in F or not torig[i]:  # if it's a feature for which I'm NOT learning a predictor

			pass

		else:

			t = copy.deepcopy(torig)
			X = copy.deepcopy(Xorig)
			Q = copy.deepcopy(Qorig)

			if type(t[i]) is not list:
				quantile_discretize([i], X, Q, t, 5)
				#sys.stderr.write("X (original):\n\n%s\n\n" % ( '\n'.join(map(lambda x : '\t'.join(map(str,x)), Xorig)) ))
				#sys.stderr.write("X (discretized column i=%d):\n\n%s\n\n" % (i, '\n'.join(map(lambda x : '\t'.join(map(str,x)), X)) ))
				#sys.stderr.write("t (original):\n\n%s\n\n" % ( torig))
				#sys.stderr.write("t (discretized column i=%d):\n\n%s\n\n" % (i, t))

			assert type(t[i]) is list # CFA only works for nominal features

			progress("Feature %i (of %i)" % (i+1, len(t)))
			verbose.flush()

			trainset=filter(lambda (x,y) : y is not None, map(lambda x : (rho[i](x),x[i]), X))
			if not trainset:
				verbose.write("# All target values are missing in the training set.\n")
				C = None
			else:
				# 2010/Nov/03:  Added try/catch here because sometimes process fails.  Notably, this happens when 
				# RIPPER is given fewer than 3 valid training labels.
				try:
					C = learn(model, types=(rho[i](t), t[i]), trainset=trainset) # "final" type-p predictor for feature i
				except Exception,ex:
					sys.stderr.write("# Failed to learn feature classifier:  %s\n" % (str(ex).strip()))
					C = None

			if not C:
				verbose.write("# No model was trained for feature=%i.  Skip this feature and compute scores based on the other features.\n" % (i+1))
				continue;
			
			observed  = map(lambda x : x[i], Q)	# list of observed x_i values (one per test example)
			predicted = list(C(map(rho[i], Q)))	# list of (predictions,distribution) 
			del C # remove model file(s), etc.

			L += 1 # valid feature 

			for q in range(len(Q)):	# for each test example
				
				(obs,(mle, P)) = (observed[q],predicted[q])

				matchcount[q] += 1 if obs == mle else 0 # what if observed[q] is missing?  No match, I guess
				
				
				try:
					total_prob[q] += P[ t[i].index(obs) ] if obs is not None else 0.0 # what if observed[q] is missing? 
				except:
					sys.stderr.write("# total_prob is %s\n" % (str(total_prob)))
					sys.stderr.write("# t[i] is %s\n" % (str(t[i])))
					sys.stderr.write("# P is %s\n" % (str(P)))
					sys.stderr.write("# obs is %s\n" % (str(obs)))
					sys.stderr.write("# q is %s\n" % (str(q)))
					total_prob[q] += P[ t[i].index(obs) ] if obs is not None else 0.0 # what if observed[q] is missing? 

				##[(si,(p,ui))] = filter(lambda (si, (p,ui)) : ui==t[i].index(obs), iwalk(reversed(sorted(gpairs(P, range(len(t[i]))))))) 
				##sys.stderr.write("# Example q=%i, t[i]=%s, P=%s, Observed (%s) is #%i (%s is #1)\n" % (q, t[i], ','.join(map(str,P)), obs, si+1, mle))

	if L==0:
		# special case:  nothing to do:
		amc = apx = [0] * len(Qorig)
	else:
		amc = map(lambda sum : sum/float(L), matchcount)
		apx = map(lambda sum : sum/float(L), total_prob)
	return list(gpairs(amc,apx))
	

# have learner learn a predictor from trainset, return predictor
def learn( learner, types, trainset ):
	trainset = list(trainset)
	predictor = learner(types, trainset)
	return predictor

# return a function f(observed,predicted) -> p(observed | predicted)
def model_distribution( ftype, A ):
	A = filter(lambda (y,g) : y is not None, A) # remove instances where observed value is missing--can't use them
	if not A:
		verbose.write("# Modeling distribution for empty sample A_i.\n")
		return ConstDistribution(1.0)
	if type(ftype) is list: # Nominal feature: Confusion Matrix
		return prob.ConfusionMatrix(values=ftype, sample=A)
	else: # Continuous feature:  Smoothed Histogram 
		return prob.ErrorDistribution( A, ftype ) # ftype = (minv, maxv)



# Convert F columns of A (in-place) to an integer, bin, 0 <= bin < bins.
# Such that the (approximately) n/bins smallest values get bin 0, and
# (approximately) all bins contain the same number of elements.
# Convert Q also, using the same thresholds used to convert A.
# Convert the types as well.
def quantile_discretize(F,A,Q,t,quantile):

	D = len(t)
	for i in F:

		if type(t[i]) is not list and t[i] is not None: # continuous feature

			# T is the set of thresholds, a bin is T[i] <= x < T[i+1].  Let
			# first threshold be -inf, so that it handles testset points < any
			# trainset point.
			T = [ -1e999 ]

			# V is the current set of non-missing feature values, not yet put
			# into any bin
			V = filter(lambda v : v is not None, gmap(lambda x : x[i], A)) 
			
			assert V # if t[i] is not None, there should be at least some non-missing values

			# it's best to choose thresholds iteratively, so that it handles
			# all distributions (e.g., zero-heavy distributions are difficult
			# to quantilicize)
			while V and len(T) < quantile: # create 'quantile' thresholds (that is, if -inf is included)
				q = quantile - len(T) + 1 # how many bins left to make
				S = list(sorted(V)) 
				q1index = int( len(S) / float(q) ) # index into S that is the 1st quantile 
				q1value = S[q1index]  
				T += [ q1value ] 
				V = filter(lambda x : x > q1value, V) # after this, only consider values that haven't yet been put into bins

			bins = len(T) # number of bins, (set to len(T) in case there are now fewer than quantile, e.g. because of fewer unique values)

			# come up with names for the labels

			L = ["x < %s"%(T[1])] + \
				map(lambda i : "%s <= x < %s" % (T[i],T[i+1]), range(1,bins-1)) + \
				["x >= %s" % (T[-1])] # corresponding labels, indicating ranges
			
			verbose.write("# Quantile discretize column %i [%s,%s] into %i bins.\n" % (i+1, t[i][0], t[i][1], bins))

			# come up with type names

			#t[i] = map(lambda bin : "bin%i" % (bin+1), range(bins)) # simple value names.
			t[i] = ["lt%s"%(T[1])] + \
			           map(lambda bin_index : "ge%slt%s" % (T[bin_index],T[bin_index+1]), range(1,bins-1)) + \
			           ["ge%s" % (T[-1])] # descriptive value names (ARFF-friendly)

			# change values of X[j][i]'s in-place
			for X in [A,Q] if A is not Q else [A]: # don't double-discretize idential sets!
				for j in range(len(X)):
					if X[j][i] is not None:
						for (bin,(theta,label)) in reversed(list(iwalk(gpairs(T,t[i])))): # go from largest threshold in order, test for x >= threshold.  This is O(n*bins) but could be O(n*lg(bins)) I suppose
							if bin==0 or X[j][i] >= theta:  
								X[j][i] = label
								break # found bin

			# Describe bins (optional)
			for (bin,(label,desc)) in iwalk( gpairs(t[i], L) ):
				Acnt = glen(gfilter(lambda x : x[i]==label, A))
				cntstr = "%d" % (Acnt)
				if A is not Q:
					Qcnt = glen(gfilter(lambda x : x[i]==label, Q))
					cntstr = "%d,%d" % (Acnt,Qcnt) 
				verbose.write("#\tbin #%d, %s, labeled %s, %s value(s).\n" % (bin+1, desc, t[i][bin], cntstr))

