# -*- coding: utf8 -*-
'''

FRaC library functions

By convension, the 'type' of nominal features is a list of strings (all
possible values) and the type of continuous features  is a tuple (minimum,
maximum value).  Thus, given feature type, programs can test if
type(feature_type)==list or type(feature_type)==tuple to determine feature
type.

By convension, missing values are indicated by a NoneType value.

Copyright © 2010 Keith Noto.  Disributed under Gnu General Public License (see http://www.gnu.org/licenses/gpl.html).

'''
__author__ = "Keith Noto <noto@cs.tufts.edu>";
__license__ = "GPL";
__copyright__ = "Copyright © 2010 Keith Noto";
VERSION = __revision__ = "1.2";
	# 1.0	2010/Apr/01	Original, with lsf module and grid search code (commented out template)
	# 1.1	2010/Jul/08	Modified to compute NS one feature at a time, then clean
	# 1.2	2010/Oct/01	Modified slightly to return a list of NS, let the caller do the summation

import sys, math, time, copy
from util import *;

verbose = open(os.devnull, 'w') # module-wide output for warnings/etc.

# Generate the normalized surprisal anomaly scores for a set of query points.
#
# X: Given normal examples--M training vectors, each of length N
# Q: Test set
# t: Types of each of N features.  Types include possible values
# F:  set of features for which to learn predictors
# rho:  rho[i] is a function reducing x_j to the vector of [potential] predictors of target feature i
# models:  A list of models that can be learned from data.  Each is a function that maps a list of (x,y) pairs to a function f(x) -> predicted y 
# folds:  cross-validation folds for evaluating the accuracy of each C_{p,i}
#
# return:  NS (normalized surprisal), a matrix such that NS[j][i] is the normalized surprisal for query example j, feature i
# 		   (this version returns NS for FRaCv1 and FRaCv2)
def ns_score(X, Q, t, F, rho, models, folds, discretize_targets):

	Xbk,Qbk,tbk = copy.deepcopy(X),copy.deepcopy(Q),copy.deepcopy(t) # backup

	t0 = time.time()
	progress = lambda msg : verbose.write("# @%.2f seconds:  %s ...\n" % (time.time()-t0, msg))

	if len(X) < folds:
		verbose.write("# Fewer training examples (%i) than folds (%i).  Doing leave-one-out.\n" % (len(X),folds))
		folds = len(X)

	# NS is normalized surprisal.
	# NS[j][i] is the normalized surprisal for query example j, feature i.
	NS1 = map(lambda i : [], range(len(Q))) # start with an empty list for each query example
	NS2 = map(lambda i : [], range(len(Q))) # start with an empty list for each query example

	for i in range(len(t)): # for each feature
			
		NS1 = map(lambda ns : ns + [0], NS1) # tenatively set NS[*][i] = 0 for all examples for this feature
		NS2 = map(lambda ns : ns + [0], NS2) # i.e., add a [0] to the end of the list

		if i in F and t[i]:  # if this is a feature for which we're learning a predictor

			if not filter(lambda x : x[i] is not None, X):
				verbose.write("# All feature #%i values missing.\n" % (i+1))
				continue

			X,Q,t = copy.deepcopy(Xbk),copy.deepcopy(Qbk),copy.deepcopy(tbk) # restore
			if discretize_targets is not None and type(t[i]) is not list:
				verbose.write("# Discretize target feature i=%d (%s)\n" % (i, t[i]))
				quantile_discretize([i], X, Q, t, discretize_targets)
			
			# entropy of the training set x_i values (a real number).  Used later to normalize surprisal
			trainset_value_entropy = prob.sample_entropy(t[i], map(lambda x : x[i], X))  ##FRaCv2## doesn't need this

			# models <- modelfactory()
			for p in range(len(models)):
					
				progress("Feature %i (of %i), supervised model %i of %i" % (i+1, len(t), p+1, len(models)))

				A = []

				learning_errors = 0
				for f in range(folds):
				
					# progress("Cross-validation fold %i of %i" % (f+1, folds))

					# divide into a train' and validation set
					(T_f, V_f) = reduce( lambda (A,B),(i,x) : (A,B+[x]) if i%folds==f else (A+[x],B), iwalk(X), ([],[]) ) 

					# set trainset to all (x,y) s.t. y is not missing (can't use those example at all)
					trainset = filter(lambda (x,y):y is not None, map(lambda x : (rho[i](x),x[i]), T_f))

					if not trainset:
						verbose.write("Feature %i, fold %i of %i:  All target values are missing in the train' set.\n" % (i+1, f+1, folds))
						continue;

					try: # this is put here since some classifiers fail some of them time due to what appear to be bugs (RIPPER)

						# learn a model, create a predictor 
						C_pif = learn(models[p], types=(rho[i](t), t[i]), trainset=trainset)
					
						# Add predictions to A_{p,i} (ignore associated probability distribution, which is part of the Predictor interface but not used)
						A += list(gpairs( map(lambda x : x[i], V_f), gmap(lambda (mle_y,_y_dist):mle_y, C_pif( map(rho[i], V_f) )) ))
						del C_pif # remove model file(s), etc.

					except Exception,ex:
						learning_errors += 1
						sys.stderr.write("\n#\n# Error (%d) on predictor evaluation fold %d:  %s\n#\n" % (learning_errors, f+1, str(ex)))

					# It is often faster to generate predictions en masse, so
					# this breaks up observations/predictions and recombines
					# them (unfortunately making it a bit messier)
		
				if not A:
					raise Exception("FRaC failed to evaluate model #%d (%s)" % (p,str(models[p])))

				elif learning_errors:
					sys.stderr.write("\n#\n# Warning:  FRaC feature predictor evaluation failed on %d of %d cross-validaion folds.\n#\n" % (learning_errors, folds))
					

				assert A # no possible reason why A would still be empty
					
				# Train final predictor
				trainset=filter(lambda (x,y) : y is not None, map(lambda x : (rho[i](x),x[i]), X)) # we already checked above for all-targets-missing

				try:
					C = learn(models[p], types=(rho[i](t), t[i]), trainset=trainset) # "final" type-p predictor for feature i

				except Exception,ex:
					# One known cause of failure is from the (WEKA) RIPPER model:
					# 	"java.lang.IllegalArgumentException: Can't normalize array. Sum is zero."
					# Not sure what causes this.
					sys.stderr.write("# Failed to train %s model for feature %d:  %s\n" % (str(models[p]), i+1, str(ex)))
					sys.stderr.write("# Skip this feature...\n")
					continue

				pr = model_distribution(t[i], A) # pr is a probabilty distribution
				
				observed  = map(lambda x : x[i], Q)			# list of observed x_i values (one per test example)
				predicted = map(lambda (y,_dist):y, C(map(rho[i], Q)))	# list of predictions (one per test example, ignore distribution)
				del C # remove model file(s), etc.

				for q in range(len(Q)):	# for each test example
					if (observed[q] is not None):	# feature value is not missing

						# Compute NS a-la FRaCv1
						NS1[q][i] = prob.surprisal(pr( observed[q], predicted[q] )) - trainset_value_entropy

						# FRaCv2
						PE = pr.entropy(predicted[q])
						S = prob.surprisal(pr( observed[q], predicted[q] ))

						if PE==0 and S==0:
							NS2[q][i] = 0
						elif PE!= 0:
							NS2[q][i] = (S - PE) / PE
						else:
							raise Exception("NS undefined")

	return (NS1,NS2)	# return NS score(s) for each test example

# have learner learn a predictor from trainset, return predictor
# written as a subroutine so we can check if all targets are equal and skip this step
def learn( learner, types, trainset ):
	trainset = list(trainset)
	values = uniq(filter(lambda y : y is not None, map(lambda (x,y):y,trainset)))
	if len(values)==1:
		
		# This turns out to be often worthwhile:  check if *all* target values
		# are identical, if so use a "parrot" predictor that just always
		# predicts that value.  Presumably, any supervised learner would do the
		# same (but less efficiently)

		# verbose.write("# Only one value (%s) present for feature, so use \"parrot\" predictor.\n" % (values[0]))
		parrot = supervised.parrot.Learner(values[0])
		predictor = parrot(types, trainset)
	
	else:
		predictor = learner(types, trainset)

	return predictor

# return a function f(observed,predicted) -> p(observed | predicted)
def model_distribution( ftype, A ):
	A = filter(lambda (y,g) : y is not None, A) # remove instances where observed value is missing--can't use them
	if not A:
		verbose.write("# Modeling distribution for empty sample A_i.\n")
		return prob.ConstDistribution(1.0)
	if type(ftype) is list: # Nominal feature: Confusion Matrix
		return prob.ConfusionMatrix(values=ftype, sample=A)
	else: # Continuous feature:  Smoothed Histogram 
		return prob.ErrorDistribution( A, ftype ) # ftype = (minv, maxv)

# Convert F columns of A (in-place) to an integer, bin, 0 <= bin < bins.
# Such that the (approximately) n/bins smallest values get bin 0, and
# (approximately) all bins contain the same number of elements.
# Convert Q also, using the same thresholds used to convert A.
# Convert the types as well.
def quantile_discretize(F,A,Q,t,quantile):

	D = len(t)
	for i in F:

		if type(t[i]) is not list and t[i] is not None: # continuous feature

			# T is the set of thresholds, a bin is T[i] <= x < T[i+1].  Let
			# first threshold be -inf, so that it handles testset points < any
			# trainset point.
			T = [ -1e999 ]

			# V is the current set of non-missing feature values, not yet put
			# into any bin
			V = filter(lambda v : v is not None, gmap(lambda x : x[i], A)) 
			
			assert V # if t[i] is not None, there should be at least some non-missing values

			# it's best to choose thresholds iteratively, so that it handles
			# all distributions (e.g., zero-heavy distributions are difficult
			# to quantilicize)
			while V and len(T) < quantile: # create 'quantile' thresholds (that is, if -inf is included)
				q = quantile - len(T) + 1 # how many bins left to make
				S = list(sorted(V)) 
				q1index = int( len(S) / float(q) ) # index into S that is the 1st quantile 
				q1value = S[q1index]  
				T += [ q1value ] 
				V = filter(lambda x : x > q1value, V) # after this, only consider values that haven't yet been put into bins

			bins = len(T) # number of bins, (set to len(T) in case there are now fewer than quantile, e.g. because of fewer unique values)

			# come up with names for the labels

			L = ["x < %s"%(T[1])] + \
				map(lambda i : "%s <= x < %s" % (T[i],T[i+1]), range(1,bins-1)) + \
				["x >= %s" % (T[-1])] # corresponding labels, indicating ranges
			
			verbose.write("# Quantile discretize column %i [%s,%s] into %i bins.\n" % (i+1, t[i][0], t[i][1], bins))

			# come up with type names

			#t[i] = map(lambda bin : "bin%i" % (bin+1), range(bins)) # simple value names.
			t[i] = ["lt%s"%(T[1])] + \
			           map(lambda bin_index : "ge%slt%s" % (T[bin_index],T[bin_index+1]), range(1,bins-1)) + \
			           ["ge%s" % (T[-1])] # descriptive value names (ARFF-friendly)

			# change values of X[j][i]'s in-place
			for X in [A,Q] if A is not Q else [A]: # don't double-discretize idential sets!
				for j in range(len(X)):
					if X[j][i] is not None:
						for (bin,(theta,label)) in reversed(list(iwalk(gpairs(T,t[i])))): # go from largest threshold in order, test for x >= threshold.  This is O(n*bins) but could be O(n*lg(bins)) I suppose
							if bin==0 or X[j][i] >= theta:  
								X[j][i] = label
								break # found bin

			# Describe bins (optional)
			for (bin,(label,desc)) in iwalk( gpairs(t[i], L) ):
				Acnt = glen(gfilter(lambda x : x[i]==label, A))
				cntstr = "%d" % (Acnt)
				if A is not Q:
					Qcnt = glen(gfilter(lambda x : x[i]==label, Q))
					cntstr = "%d,%d" % (Acnt,Qcnt) 
				verbose.write("#\tbin #%d, %s, labeled %s, %s value(s).\n" % (bin+1, desc, t[i][bin], cntstr))

