#!/usr/bin/env python
# -*- coding: utf8 -*-
'''

Functions for inferring data types from file.  This is simple:  Types are one of:

(i) continuous, consisting of all numeric values (convert these to float or int)
(ii) nominal, consisting of not all numeric types (convert these to str)
(iii) None, meaning that the feature type is unkown, or sometimes that the feature is to be dismissed (e.g., perhaps because all values are identical)

Copyright © 2010 Keith Noto.  Disributed under Gnu General Public License (see http://www.gnu.org/licenses/gpl.html).
'''
__author__ = "Keith Noto <noto@cs.tufts.edu>";
__license__ = "GPL";
__copyright__ = "Copyright © 2010 Keith Noto";
VERSION = __revision__ = "1.1"
# 1.1	2010/Oct	Overhaul

import sys,string,StringIO
from frac.util import *

# Define symbols in the data we interpret as missing
def is_missing(symbol):
	return symbol is None or \
	       symbol is type(None) or \
		   str(symbol).lower() in ['nan','','?']

# Automatically cast numbers to floats, and missing data to NoneType.  This
# helps reduce memory requirements and make features "uniform" 
# (3.5 and 3.50 both mean the same)
def typecast(v):
	if is_missing(v):
		return None
	elif type(v) in [float,int,long]:
		return v # already numeric
	elif type(v) is str:
		try:
			return int(v) # cast numbers with no decimal to integer type
		except ValueError:
			try:
				return float(v) # cast anything interpretable as a number to a real
			except:
				return v # it's just a string, then
	else:
		return v


# Automatically infer types for each column.
# The types generated by this function are either 
# (i) a pair, (min value, max value) for numeric types, or 
# (ii) a list of possible values for non-numeric types, or (rarely)
# (iii) None for columns where ALL values are missing
def infer_types(X):
	X=list(X)
	if X:
		N = len(X[0])
		for i in range(N):
			values = {}
			continuous = True
			minv,maxv = 1e999,-1e999;
			for value in map(typecast, map(lambda x : x[i], X)):
				if value is not None:
					values[value] = True # do this every record in case we find out later that feature is nominal
					if type(value) in [float,int,long]:
						minv = min(value,minv)
						maxv = max(value,maxv)
					else:
						continuous = False
			yield None if not values else (minv,maxv) if (continuous and len(values) > 2) else sorted(values.keys())
			# note:  continuous features with two values (0,1; -1,+1; 1,2; etc.) are considered to be binary.

# Decide which types are "usable," which depends on the data.  
# Generate usable types and assign NoneType to the rest.
# Issue warnings explaining all this.
def usable_types(X, t, warnings):
	for (i, t_i) in iwalk(t):
		if t_i is None:
			warnings.write("# Warning:  Cannot infer type of column %i.\n" % (i+1))
			yield None
		elif not filter(lambda x : x[i] is not None, X):
			warnings.write("# Warning:  Column %i is missing all data.\n" % (i+1))  # infer_types would catch this, but maybe column types are user-defined
			yield None
		elif type(t_i) is list and len(t_i)==1  or  type(t_i) is tuple and len(t_i)==2 and t_i[0]==t_i[1]:
			warnings.write("# Warning:  There is only one value in column %s (%s).\n" % (i+1, t_i[0]))
			yield None
		elif len(t_i)==glen(gfilter(lambda x : not is_missing(x[i]), X)):
			warnings.write("# Warning:  All values in column %s are different (%s, etc.)  Are these just an example IDs?\n" % (i+1, ', '.join(t_i[:2])))
			yield None
		else:
			yield t_i

# generate types from metadata
def parse_metadata(input):
	t = []
	lineno = 0
	for line in input:
		lineno += 1
		try:
			if len(line.split()) == 2:
				[ column, typedecl, values ] = line.split(None) + [ '' ]
			else:
				[ column, typedecl, values ] = line.split(None, 2) 
			i = int(column)-1	# convert to zero-origin
			if i >= len(t):
				t = t + [None] * (i + 1 - len(t)) # caller may not know how many features
			if t[i] is not None:
				sys.stderr.write("# Warning:  Multiple definitions of column %s.\n" % (column))
			typename = typedecl.strip().lower()
			values = map(string.strip,values.split(','))

			# accept these types as 'ignore this feature for whatever reason'
			if typename.lower() in ['ignore', 'const', 'constant', 'id', 'identifier', ]:
				t[i] = None
			
			# accept these types as a nominal feature
			elif typename.lower() in ['nominal', 'binary', 'boolean', 'categorical', 'hierarchical',]:
				t[i] = list(sorted(map(typecast,uniq(values))))

			# accept these types as a continuous feature 
			elif typename.lower() in ['continuous', 'numerical', 'linear', 'real',]:
				t[i] = tuple(map(float,values))

			# Note that the 'ordered' feature type is tricky, because
			# classifiers don't always handle ordered features or targets, and
			# some have no way of doing it.  FRaC doesn't handle them
			# explicitly:  Users must specify them them nominal or continuous
			# (and in some cases, update the input data to match).

		except Exception,ex:	
			raise(Exception("Error parsing Meta-Data.  Line %i was:  %s" % (lineno,line[:-1])))
	return t

# generate lines of metadata:  each is (column number, type definition, values)
def metadata(t):
	for (i, t_i) in iwalk(t):
		column = i+1 # to print in 1-origin 
		typedef = "%s" % ("ignore    " if t_i is None else \
		                  "nominal   " if type(t_i) is list else \
		                  "continuous") # column type
		values = ','.join(map(str,[] if t_i is None else t_i))
		yield (column, typedef, values)


# If called as a program, print metadata for input matrix
def main():

	import sys,string,optparse,os.path,signal;
	signal.signal(signal.SIGPIPE, signal.SIG_DFL); # properly handle SIGPIPE when piping output 

	PROGRAM = os.path.basename(sys.argv[0]); 
	USAGE = "[options] < input";
	VERSION = "1.0";
	SYNOPSIS = "Infer column types from tabular data.  Disributed under Gnu General Public License (see http://www.gnu.org/licenses/gpl.html).";

	parser = optparse.OptionParser(usage="%s %s"%(PROGRAM,USAGE), version="%s version %s"%(PROGRAM,VERSION), description=SYNOPSIS);
	parser.add_option("-d", "-F", "--delimiter",   dest="delim", type="string", metavar="String",  default=None, action="store", help="Field delimiter (default is any white space)")
	(options, args) = parser.parse_args();

	write_metadata( infer_types( map(lambda x : map(typecast, x), \
	                             map(lambda x : map(string.strip, x), \
								 map(lambda line : line.split(options.delim), \
								 sys.stdin if not args or args[0].strip()=='-' else open(args[0],'r'))))), \
				    sys.stdout )


	return;

if (__name__ == "__main__"):
	main();

