#!/usr/bin/env python
# -*- coding: utf8 -*-
'''
Modeling probability distributions.
Copyright © 2009.  Disributed under Gnu General Public License (see http://www.gnu.org/licenses/gpl.html).

'''
__author__ = "Keith Noto <noto@cs.tufts.edu>"
__license__ = "GPL"
__copyright__ = "Copyright © 2009 Keith Noto"
VERSION = __revision__ = "1.1" 

import sys,string,random,math
inf = float('inf')

# sample mean
def mean(X):
	( sum, N ) = reduce(lambda (a,N),x : (a+x,N+1) , X, (0.0,0));
	return sum/N;

# sample standard deviation
def stdev(X, mu=None):
	if mu is None:
		X = list(X)
		mu = mean(X) # pass #1
	( ssd, N ) = reduce(lambda (a,N),x : (a+((mu-x)*(mu-x)) , N+1), X, (0.0,0))
	return math.sqrt( ssd / N )

# log base 2
def lg(x):
	return math.log(x) / 0.69314718055994528622676398299518041312694549560546875; # log(2)

# Self-information
def surprisal(p):
	return inf if p==0 else -lg(p)

# Entropy of a multinomial
def entropy(D):
	return sum(map(lambda p : 0 if p==0 else -p*lg(p), D))

# Simple (as can be) representation of a Gaussian distribution
class Gaussian:
	def __init__(self,mu=0.0,sigma=1.0):
		self.mu = float(mu)
		self.sigma = float(sigma)
	def p(g,x):
		# 1 / sqrt(2PI) = 0.3989422804014327 
		d = x - g.mu
		return (1/g.sigma) * 0.3989422804014327 * math.exp( -d*d / (g.sigma*g.sigma) )
	def CDF(g,x):
		# From http://en.wikipedia.org/wiki/Gaussian_Distribution
		return 0.5 * ( 1 + erf( (x - g.mu) / math.sqrt(2*g.sigma*g.sigma) ) )

# A weighted mixture of PDFs.
class Mixture:
	def __init__(self, mixture={}):
		self.mixture=mixture;	# { component:weight }
	#def __str__(m):
	#	return "{%s}" % ('; '.join(map(lambda (c,w) : "%s, weight=%s" % (str(c),w), m.mixture.iteritems())))
	def p(m,x):
		return sum(map(lambda (c,w) : w*c.p(x), m.mixture.iteritems()))
	def CDF(m,x):
		return sum(map(lambda (c,w) : w*c.CDF(x), m.mixture.iteritems()))
	#def mean(m):
	#	# mean is always weighted mean (?)
	#	sum = 0.0
	#	mass = 0.0
	#	for (c,weight) in m.mixture.iteritems():
	#		sum += weight * c.mean()
	#		mass += weight;	# I'm mixing my metaphors
	#	return sum / mass;	# mass should be 1

# From The Endeavor:  http://www.johndcook.com/blog/2009/01/19/stand-alone-error-function-erf
# Here's a Python implementation of erf(x) based on formula 7.1.26 from A&S. The maximum error is below 1.5 x 10-7.
#
# Handbook of Mathematical Functions: with Formulas, Graphs, and Mathematical Tables (Paperback)
# Milton Abramowitz (Editor), Irene A. Stegun (Editor) 
# 
# http://www.amazon.com/gp/product/0486612724?ie=UTF8&tag=theende-20&linkCode=xm2&camp=1789&creativeASIN=0486612724
#
def erf(x):
	# constants
	a1 =  0.254829592
	a2 = -0.284496736
	a3 =  1.421413741
	a4 = -1.453152027
	a5 =  1.061405429
	p  =  0.3275911

	# Save the sign of x
	sign = 1
	if x < 0:
		sign = -1
	x = abs(x)

	# A & S 7.1.26
	t = 1.0/(1.0 + p*x)
	y = 1.0 - (((((a5*t + a4)*t) + a3)*t + a2)*t + a1)*t*math.exp(-x*x)

	return sign*y

# Only integer values 0,1,...,n-1 are applicable
# Can be used as a PDF but with care...
class Multinomial:	
	def __init__(self,d):
		self.d = list(d);	# use 'd' the simple representation, list of probabilities
	def __str__(self):
		return "Multinomial PDF {%s}" % ( ','.join(map(lambda i : "%i:%g"%(i,self.d[i]), range(len(self.d)))))
	def normalize(self):
		Z = float(sum(self.d))
		self.d = map(lambda p : p/Z, self.d);	 # assume sum(d) > 0
	def p(self,i):
		if not ( 0 <= i < len(self.d) ):
			raise Exception("Illegal domain %s for multinomial(%s)" % (i, len(self.d)))
		return self.d[int(i)]
	def sample(self):
		mass = sum(self.d)
		for i in range(len(self.d)):
			if random.random() * mass < self.d[i]:
				return i
			mass -= self.d[i]
	def mean(self):
		return sum(map(lambda i : i * self.d[i], range(len(self.d)))) / sum(self.d)
	def median(self):
		t = sum(self.d)/2 # i.e. 0.5
		cd = 0	# cum. density (of the bins counted so far)
		for i in range(len(self.d)):
			if cd+self.d[i] >= t:
				return i + (t-cd)/self.d[i]
			cd += self.d[i]
	def mode(self):
		ans = 0
		for i in range(1,len(self.d)):
			if self.d[i] > self.d[ans]:
				ans = i
		return ans
	def CDF(self,i):
		return sum(map(lambda j : self.d[j], range(i+1)))
	def inverseCDF(self,p):
		Z = sum(self.d)
		t = p * Z
		cd = 0
		for i in range(len(self.d)):
			cd += self.d[i]
			if cd >= t:
				return i
	def variance(self):
		mu = self.mean()
		xssd = 0
		for i in range(len(self.d)):
			xssd += self.d[i] * (mu-i) * (mu-i)
		return xssd 
	def __iter__(self):
		return iter(self.d)

# Smooth a multinomial (list of probabilities) Gaussian-shaped kernel
#
# This has the side-effect of making the distribution look more normal (i.e.
# the middle bins get more mass than the outer ones)
#
#
def smooth(A, sigma=1):
	n = len(A)	
	B = [0] * n
	for i in range(n):
		g = Gaussian(mu = i+0.5, sigma=sigma)
		mass = g.CDF(n) - g.CDF(0)
		for j in range(n):
			# transfer mass from A[i] to B[j]
			B[j] += A[i] * (g.CDF(j+1) - g.CDF(j)) / mass
	return B

# Matrix is mapping from predicted value -> multinomial over observed values
class ConfusionMatrix:
	def __init__(self, values, sample, pseudocount=1):
		self.matrix = dict(map(lambda v : (v, dict(map(lambda v : (v, pseudocount), values))), values))
		for (h, g) in sample:
			self.matrix[ g ][ h ] += 1
		for g in self.matrix:
			# normalize matrix[g]
			Z = float(sum(map(lambda (h,count) : count, self.matrix[g].iteritems())))
			for h in self.matrix[g]:
				self.matrix[g][h] /= Z
	def __call__(self, observed, predicted): # compute p(observed | predicted)
		return 0.0 \
			if predicted not in self.matrix or observed not in self.matrix[predicted] \
			else self.matrix[predicted][observed]
	def entropy(self, prediction):
		return entropy(self.matrix[prediction].values())

# maps values to bin space
class SmoothedHistogram: 
	def __init__(self, sample, (minv,maxv)):

		# Step 1 is to populate the histogram (a list of real-numbered
		# probability values), and step 2 is to smooth it (See `smooth'
		# function).  We'll actually allocate room for 3x the number of bins,
		# since the smoothing process may assign nonzero probability to some
		# bins outside the default range.  

		sample = list(sample) # in case sample is generator--we'll need a few passes over it
		assert all(map(lambda v : minv <= v <= maxv, sample)) 

		self.epsilon=1e-99 # tiny value to ensure no zero probabilities
		n = len(sample)
		self.bins = 3 * max(2, int(math.sqrt(n))) # number of bins proportional to sqrt(n), but add 3x room for runoff from smoothing
		vrange = maxv-minv 
		self.minv, self.maxv = minv - vrange - 0.001, maxv + vrange + 0.001 # expand range also to account for runoff from smoothing.  That is, minv is supposed to be the smallest legal value, but self.minv now refers to the smallest value that actually has a corresponding bin.  Add a fudge factor to account for minv==maxv and int truncation.
		self.H = [0]*self.bins	# the histogram
		for x in sample:
			self.H[ self.map(x) ] += 1
		self.H = map(lambda count : count/float(len(sample)), self.H) # normalize
		self.H = smooth(self.H, sigma = 1) # sigma=1 bin

	def map(self, x):
		return int(self.bins * (x - self.minv) / (self.maxv - self.minv)) # x maps to this bin

	def __call__(self, x):
		bin = self.map(x)
		return max(self.H[bin],self.epsilon) if 0 <= bin < len(self.H) else self.epsilon # return tiny epsilon if probability goes to zero or if out of range

class ErrorDistribution:
	def __init__(self, sample, (minv,maxv)):
		err_sample = map(lambda (y,g) : y-g, sample)
		self.smoothhist = SmoothedHistogram( err_sample, (min(err_sample),max(err_sample)) )
	def __call__(self, observed, predicted): # compute p(observed | predicted)
		return self.smoothhist( observed - predicted )
	def entropy(self, prediction):
		return entropy(self.smoothhist.H)

class ConstDistribution:
	def __init__(self, p):
		self.p = p
	def __call__(self, observed, predicted): # compute p(observed | predicted)
		return self.p
	def entropy(self, prediction):
		return 0

def sample_entropy( ftype, sample, pseudocount=1 ): # (pseudocount for nominal types only)
	sample = filter(lambda y : y is not None, sample)   # remove missing examples
	if not sample:
		raise Exception("Empty distribution")
	if type(ftype) is list: # nominal feature
		M = dict(map(lambda v : (v,pseudocount), ftype))
		for v in sample:
			M[v] +=1
		Z = float(sum(map(lambda (v,count):count, M.iteritems())))
		return entropy( map(lambda (v,count) : count/Z, M.iteritems()) )
	else:   # continuous feature
		return entropy( SmoothedHistogram(sample, ftype).H )

