#!/usr/bin/env python
# -*- coding: utf8 -*-
'''

Utilities for learning with SVMs (using LIBSVM).  This includes code to do a
parameter grid search to find the best parameter settings, but it may be
commented out in the online downloadable version, so you have to update the
code to do the grid search you want.

Copyright © 2010 Keith Noto.  Disributed under Gnu General Public License (see http://www.gnu.org/licenses/gpl.html).

'''
__author__ = "Keith Noto <noto@cs.tufts.edu>"
__license__ = "GPL"
__copyright__ = "Copyright © 2010 Keith Noto"
VERSION = __revision__ = "1.1"
# 1.0	2010    	Original
# 1.1	2010/Sep	Gridsearch(es)

import sys,os,string,time,tempfile,math,hashlib,re,subprocess,tempfile
import parrot,lsf # may or may not be used
import storage
from util import *

# module parameters:
verbose = open(os.devnull, 'w') # general messages, process output, etc.
gridout = sys.stderr            # grid search progress

class Predictor:
	# pcmd: predict shell command (e.g., 'svm-predict')
	# svmfile: stored model path/filename
	# ftypes: list of feature types, in order (types are list for nomainal features, tuple for continuous)
	# ttype: type of target
	# clean:  delete model file on del?
	# pid:  wait for this process ID (it's the training process) before making predictions
	def __init__(self,pcmd,svmfile,ftypes,ttype,clean,pid):
		(self.pcmd, self.svmfile, self.clean, self.ftypes, self.ttype, self.pid) = (pcmd, svmfile, clean, ftypes, ttype, pid)
	
	# generate (prediction, distribution-when-available) pairs for each x in Q
	def __call__(self,Q):

		if self.pid:
			# must wait for training process to finish, before I can make the predictions
			os.waitpid(self.pid,0)	
			self.pid = None # train once / predict many times

		Q = list(numericize(Q, self.ftypes))	 # translate to SVM-friendly format

		predictions = list(interpret(predict(Q, self.pcmd, self.svmfile), self.ttype))
		
		if len(predictions) != len(Q):
			raise Exception("\n\nNumber of SVM predictions (%i) not equal to number of test set examples (%i)." % (len(predictions),len(Q)))

		for prediction in predictions:
			yield prediction

	def __del__(self):
		if self.clean:
			rm(self.svmfile)


# Learner just needs to know the train/predict commands.  We assume models are
# trained and predictions are made, respectively, with these shell commands:
#
# $ <train command> <trainset data file> <model output file>
# $ <predict command> <test set data file> <model file> <predictions output file>
#
# This is the case for LIBSVM and SVM-Light
#
class Learner:

	def classifier():
		return False # SVM is a regression model, too

	# r_tcmd: train command for regression (e.g., 'svm-train -s 3 -t 1 -d 2 -c 1')
	# r_pcmd: predict command for regression (e.g., 'svm-predict')
	# c_tcmd: train command for classification
	# c_pcmd: predict command for classification
	# clean:  delete model files when done?
	def __init__(self, r_tcmd, r_pcmd, c_tcmd, c_pcmd, clean=True):
		self.r_tcmd, self.r_pcmd, = r_tcmd, r_pcmd	# for regression
		self.c_tcmd, self.c_pcmd, = c_tcmd, c_pcmd  # for classification
		self.clean = clean

	# ftypes:  feature types
	# ttype: target type
	# T: trainset; list of (x, y) pairs where y is target value, x is feature vector
	def __call__(self, (ftypes,ttype), T):

		T = list(filter(lambda (x,y) : y is not None, T))

		if not T:
			raise(Exception("\n\nCall to SVM Learner with zero labeled training examples"))
			# Note:  One could accept this scenerio as one that probably
			# doesn't affect FRaC output by returning
			# parrot.Predictor(ttype[0]), which arbitrarily always predicts the
			# first label / minimum value.

		# reformat training data
		T_x = numericize(map(lambda (x,y):x, T), ftypes) # numericize:  deterministically replace features with numerical values
		T_y = map(lambda (x,y):y, T)
		if type(ttype) is list: # (nominal)
			T_y = map(lambda y : ttype.index(y), T_y)	# map targets to 0,1,2,... indecies in ftypes list
			nttype = range(len(ttype)) # rather than re-numericize the same data over and over, just remember the "numericized" target type 
		else:
			nttype = ttype # continuous feature target type is already numericized
		T2 = list(gpairs(T_x, T_y))

		# decide on training,predicting command
		(tcmd,pcmd) = (self.c_tcmd, self.c_pcmd) if type(ttype) is list else (self.r_tcmd,self.r_pcmd)

		# begin gridsearch code
		'''
		# One may append other parameter settings to the SVM training command
		# by doing a parameter sweep.  There are many ways to do this and the
		# following is just one *EXAMPLE* 

		origin = [] # origin point in parameter settings space
		objective = Accuracy()	# 1.0 minus Zero-one loss
		cmp = (lambda a,b : -1 if a > b else 1 if a < b else 0) # cmp is saying that higher scores are better

		if re.search('-t\s*2', tcmd) and not re.search('-g', tcmd): # RBF w/o given gamma value
			origin += [ ('-g', 0.1) ] # add gamma then

		if not re.search('-c',tcmd):
			origin += [ ('-c', 1.0) ] # the cost parameter 

		if type(ttype) is not list: # continuous:  add epsilon parameter and change objective 
			if not re.search('-p',tcmd):
				origin += [ ('-p', 0.1) ]
			objective = MSE()	 # Mean Squared Error
			cmp = (lambda a,b : -1 if a < b else 1 if a > b else 0)  # this says smaller is better, for MSE

		folds = len(T2) if len(T2) < 5 else 5	# cross-validation to estimate performance of parameter settings

		parameters = axisclimb(tcmd, pcmd, origin, dataset=T2, ftypes=ftypes, ttype=nttype, folds=folds, objective=objective, cmp=cmp, base=10, radius=1)
		tcmd = "%s %s" % (tcmd, ' '.join(map(lambda (flag,value) : "%s %s" % (flag,value), parameters)))
		gridout.write("# Result of parameter grid search:  %s\n" % (tcmd))
		gridout.flush()
		'''
		# end gridsearch code

		svmfile = name_svmfile(T, tcmd)
		pid = train(T2, tcmd, svmfile)
		return Predictor(pcmd, svmfile, ftypes=ftypes, ttype=ttype, clean=self.clean, pid=pid)


# Numericize:  Deterministically replace nominal features with numerical ones
# For k-valued nominal features, create k binary features.
# Replace missing values in continuous features with a mid-range value
def numericize(X, t):
	t = list(t)
	for x in X:
		# map x -> x2, where x2 has the same data in numercial features
		assert len(x)==len(t)	# x (data vector) and t (type of each component) are parallel
		yield list(numericize_row(x, t))
def numericize_row(x, t):
	for i in range(len(t)):
		if type(t[i]) is list: # nominal, create binary feature for each possible value
			# It would be nice to keep binary features as just one column, but
			# that wouldn't handle missing data
			for v in t[i]:  # for each possible value v, 
				yield 1 if x[i]==v else 0	 # (handles missing x[i], too--set all values to zero)
		else: # continuous
			if x[i] is None:
				# missing data, use a mid-range value
				(minv,maxv) = t[i] # types of continuous features is a (min,max) tuple, so this should always work
				yield 0 if minv==-maxv else ((minv+maxv)/2.0) # handles (-inf,+inf)
			else:
				yield x[i] # not missing--just use value itself

# give svm files a standard name, based on the training set and command
# (so they can be recalled if the exact same job is to be run again)
def name_svmfile(trainset, tcmd):
	svmcode = hash(str(trainset)+str(tcmd))
	svmfile = "%s/%s.svm" % ( chkdir(storage.cache), svmcode )
	return svmfile

# Train a model.  @return process ID to wait for.
def train(trainset, tcmd, svmfile):

	fork = False # should training process fork?  This is useful when multiple processors are available.

	if os.path.exists(svmfile) and os.path.getsize(svmfile):

		verbose.write("# Use existing model %s\n" % (svmfile))
		return None	# no pid

	else:

		pid = os.fork() if fork else None	# available on Unix 
		
		if not pid:

			trainfile = tempfile.mktemp('.smx.tmp', 'svm.trainset.', chkdir(storage.tmp))
			write(trainset, open(trainfile,'w'))
			cmd = "%s %s %s" % (tcmd, trainfile, svmfile)
			
			verbose.write("# %s:\n#\t$ %s\n" % (time.strftime("%A %B %d %I:%M:%S %p %Y"), cmd));  # echo command

			## Option 1:  Just run the process.
			#p = subprocess.Popen([cmd], shell=True, stdin=None, stdout=verbose, stderr=None, close_fds=True)
			#p.wait()

			## Option 2:  changing the convergence criteria if the job takes too long
			# this function runs the svm-train command for a while, and if it
			# doesn't finish, it kills it, and restarts another with a greater
			# value for convergence tolerance criterion.
			svm_train_with_increasing_tolerance(cmd, timeout=60)  # timeout given in seconds
			
			## Option 3:  submit LSF job
			# LSF submission should be run with fork, since lsf.submit will block
			# lsf.submit(cmd, jobname=hash(str(trainset)+tcmd+svmfile), jobdir=cache)
		
			if not os.path.exists(svmfile):
				raise Exception("\n\nLIBSVM process failed:\n\n%s\n\n" % (cmd)); 

			rm(trainfile)
			if fork:
				sys.exit(0)
		return pid

# Run a train process, (i.e. "svm-train ...") but increase the epsilon
# tolerance if the process is taking too long.  This 
def svm_train_with_increasing_tolerance(cmd, timeout, initial_tolerance = 0.01, factor = 10):

	e = initial_tolerance
			
	if not re.search('svm-train',cmd) and not re.search('liblinear',cmd): # added 'liblinear' to handle special case
		raise Exception("Assertion Failed, that LIBSVM train command \"%s\" would have 'svm-train' in it."  % (cmd)) # we're using LIBSVM for this

	if re.search('-e[\s\.0-9\-]', cmd):
		verbose.write("# Warning:  Call to `svm_train_with_increasing_tolerance,' but tolerance is given: %s\n" % (cmd))
		p = subprocess.Popen([cmd], shell=True, stdin=None, stdout=verbose, stderr=None, close_fds=True)
		p.wait()

	else:
		
		while True:

			#e_cmd = cmd.replace('svm-train', 'svm-train -e %g ' % (e))
			mo = re.search('[^\\\\]\\s', cmd)
			prog, args = cmd[:mo.end()], cmd[mo.end():]
			e_cmd = "%s -e %g %s" % (prog, e, args)

			verbose.write("# $ %s\n" % (e_cmd))
		
			p = subprocess.Popen([e_cmd], shell=True, stdin=None, stdout=verbose, stderr=None, close_fds=True)

			sleep = 0.000001 # seconds; 0.001 = 1 ms, 0.000001 = 1us
			elapsed = 0
			while elapsed < timeout and p.poll() is None:
				time.sleep(sleep)
				elapsed += sleep
				sleep = min( sleep*2, timeout-elapsed, 60 ) # wait a little longer for process

			if p.poll() is None:
				verbose.write("# KILL train process (taking > %s seconds): %s\n# Re-run with greater tolerance of termination criterion.\n" % (timeout, e_cmd))
				p.terminate()
				e *= factor
			else:
				break # p.poll() is not None => process finished => quit loop


# Run classification process; generate output
def predict(testset, pcmd, svmfile):

	testset = map(lambda x : (x,float()), testset); # add dummy targets for file format
	testfile = tempfile.mktemp(".smx.tmp","svm.testset.",storage.tmp)
	outputfile = tempfile.mktemp(".out.tmp","svm.output.",storage.tmp)
	write(testset,open(testfile,'w'))

	cmd = "%s %s %s %s" % (pcmd,testfile,svmfile,outputfile)
	verbose.write("# %s:\n#\t$ %s\n" % (time.strftime("%A %B %d %I:%M:%S %p %Y"), cmd));  # echo command

	p = subprocess.Popen([cmd], shell=True, stdin=None, stdout=verbose, stderr=None, close_fds=True)
	p.wait()

	if not os.path.exists(outputfile):
		raise Exception("\n\nSVM process failed: %s" % (cmd))

	outputfile_stream = open(outputfile,'r')

	# loop one ahead of the line we're generating, since we want to be sure to
	# close the file (iteration may not reach StopIteration since user will
	# know how many predictions are generated)
	queue = None
	for line in outputfile_stream:
		if queue:
			yield queue
		queue = line
	outputfile_stream.close() # close the file before yielding the last prediction
	rm(outputfile)
	rm(testfile)
	yield queue # final prediction

# standard interpretation of predict output
def interpret(output, ttype):

	labels = []
	for line in output:

		split = line.split()

		if line.find('label') >= 0:
			# special first line for Platt-scaled probability output
			labels = map(int, split[1:])

		elif labels:
			assert type(ttype) is list # labels must mean nominal target type

			# assign proper labels
			distribution = [0]*len(ttype)
			for (i,p) in gpairs(labels,map(float,split[1:])):
				distribution[i]=p
			prediction = ttype[int(split[0])] # also distribution.index(max(distribution))

			yield (prediction,distribution)

		else:
			prediction = ttype[int(split[0])] if type(ttype) is list else float(split[0])
			yield (prediction, None) # one prediction per line

# write X in libsvm-format to stream
# X:  list of (sparse vector,target)
def write(X, out):
	for (x,y) in X:
		out.write("%g\t%s\n" % (y, ' '.join(map(lambda i : "%i:%g" % (i+1, x[i]), range(len(x))))))

# Remove a temporary file (may be disabled for debugging)
def rm(filename):
	#sys.stderr.write("# Warning:  Keeping temporary file %s\n" % (filename))
	os.unlink(filename)

# Codify whatever this is
def hash(thingy):
	# if hashlib is not available, write another hash function
	return hashlib.md5(str(thingy)).hexdigest()

# The following code is for the parameter gridsearch.
#
# We consider SVM classifiers to be "ugly" if they don't first choose parameter
# settings in a disciplined way; e.g., a data-driven grid search over
# combinations of parameter values.  
#

# The following code does a gridsearch by trying each possible combination of a
# range of candidate parameter values.   This is often too time-consuming for a
# single processor, but this code can take advantage of mulitple processors or
# a cluster submission (see lsf.py)
#
# @param tcmd		The base of the command, e.g. "svm-train -t 1 "
# @param pcmd		The prediction command, e.g. "svm-predict"
# @param points     Points to try a list of (flag, values); e.g., [ ('-c', [0,0.1]), ('-g', [0.1,0.01]) ]
# @param trainset	set of (x,y) pairs (x is the feature vector, y is the target)
# @param folds	   number of cross-validation folds
# @param objective	function mapping (observed,predicted) pairs to some real-valued score
# @param cmp		  function takes two arguments, returns -1 if the first is a better objective score
# @return           best parameter settings; e.g., [ ('-c', 0.1), ('-g', 0.01), ('-p', 0) ]
def gridsearch(tcmd, pcmd, points, dataset, ftypes, ttype, folds, objective, cmp):

	flags = map(lambda (flag,values) : flag, points)

	# this will evaluate grid search points
	evaluate = Evaluator(tcmd,pcmd,flags,list(dataset),ftypes,ttype,folds,objective,cmp) 

	# step 1:  Try all grid search points
	pdata = []	# list of (training command, predictors-for-all-folds)
	for point in combinations(map(lambda (flag,values) : values, points)):
		pdata += [ evaluate.train(point) ] # starts process running, adds pid and other data to this list

	# step 2:  evaluate all of them
	best_point = None
	for (point,tuples) in pdata:
		score = evaluate.eval(point, tuples)
		best_point = evaluate.best( point )

	return list(gpairs(flags, best_point)) # reassign flags

# Do a hill-climbing log-scale grid search
#
# @param tcmd		The base of the command, e.g. "svm-train -t 1 "
# @param pcmd		The prediction command, e.g. "svm-predict"
# @param origin	    The search origin; e.g. [ ('-c',1), ('-g', 0.1), ('-p', 0.1) ]
# @param trainset	set of (x,y) pairs (x is the feature vector, y is the target)
# @param folds	    number of cross-validation folds
# @param objective	function mapping (observed,predicted) pairs to some real-valued score
# @param cmp		function takes two arguments, returns -1 if the first is a better objective score
# @param base	    base for log-scale grid search
# @return           best parameter settings; e.g., [ ('-c', 0.1), ('-g', 0.01), ('-p', 0) ]
def hillclimb(tcmd, pcmd, origin, dataset, ftypes, ttype, folds, objective, cmp, base):
	
	log    = lambda point : map( lambda c : round(math.log(c,base), 10), point ) # map a point to log-scale (the conversions often cause rounding errors, and we don't want to re-evaluate essentially the same parameter combinations) 
	linear = lambda point : map( lambda c : round(math.pow(base,c), 10), point ) # map a point back to linear-scale
	neighbors = lambda origin : gmap(linear,eight_neighbors(log(origin), radius=1)) # define 'neighbors' function

	flags = map(lambda (flag,value) : flag, origin)		# the given flags
	origin = map(lambda (flag,value) : value, origin)	# just the point

	# Create an point-in-parameter-space evaluator
	evaluate = Evaluator(tcmd,pcmd,flags,list(dataset),ftypes,ttype,folds,objective,cmp) 

	previous_origin = None

	while previous_origin != origin:

		previous_origin = origin
	
		# step 1:  try all neighbors
		pdata = []
		for point in neighbors(origin):	# find neighbors in log space, yo
			pdata += [ evaluate.train(point) ] # starts process running, adds pid and other data to this list
	
		# step 2:  evaluate all of them
		for (point,tuples) in pdata:
			score = evaluate.eval(point, tuples)
			origin = evaluate.best( point )

	return list(gpairs(flags, origin)) # reassign flags


# Do a single-parameter-at-a-time-hill-climbing log-scale grid search
#
# @param tcmd		The base of the command, e.g. "svm-train -t 1 "
# @param pcmd		The prediction command, e.g. "svm-predict"
# @param origin	    The search origin; e.g. [ ('-c',1), ('-g', 0.1), ('-p', 0.1) ]
#                   Parameter values are explored in this order  
# @param trainset	set of (x,y) pairs (x is the feature vector, y is the target)
# @param folds	    number of cross-validation folds
# @param objective	function mapping (observed,predicted) pairs to some real-valued score
# @param cmp		function takes two arguments, returns -1 if the first is a better objective score
# @param base	    base for log-scale grid search
# @param radius     how far along each axis to look
# @return           best parameter settings; e.g., [ ('-c', 0.1), ('-g', 0.01), ('-p', 0) ]
def axisclimb(tcmd, pcmd, origin, dataset, ftypes, ttype, folds, objective, cmp, base, radius):

	log    = lambda point : map( lambda c : round(math.log(c,base), 10), point ) # map a point to log-scale (the conversions often cause rounding errors, and we don't want to re-evaluate essentially the same parameter combinations) 
	linear = lambda point : map( lambda c : round(math.pow(base,c), 10), point ) # map a point back to linear-scale
	neighbors = lambda c : linear(gmap(lambda singleton:singleton[0], eight_neighbors(log([c]),radius=radius)))	# which neighbors of a given value?
	
	flags = map(lambda (flag,value) : flag, origin)		# the given flags
	origin = map(lambda (flag,value) : value, origin)	# just the point
		
	# Create an point-in-parameter-space evaluator
	evaluate = Evaluator(tcmd,pcmd,flags,list(dataset),ftypes,ttype,folds,objective,cmp) 

	previous_origin = None
	while previous_origin != origin:
		previous_origin = origin
		for index in range(len(origin)): # which parameter?
			# step 1 try points along the line 
			pdata = []
			for value in neighbors(origin[index]):
				pdata += [ evaluate.train( origin[:index] + [value] + origin[(index+1):] ) ]
			# step 2 evaluate all
			for (point,tuples) in pdata:
				score = evaluate.eval(point, tuples)
				origin = evaluate.best( point )
	return list(gpairs(flags, origin)) # reassign flags



# SVM Parameter evaluator
class Evaluator:
	
	def __init__(self, tcmd,pcmd,flags,dataset,ftypes,ttype,folds,objective,cmp):
		self.tcmd,self.pcmd,self.flags,self.dataset,self.ftypes,self.ttype,self.folds,self.objective,self.cmp = tcmd,pcmd,flags,dataset,ftypes,ttype,folds,objective,cmp
		self.cache={}	# which parameter combinations have already been tested (useful for heuristic searches so we keep track here if any points are re-evaluated)
		self.printcache={} # which model scores have been printed?
		self.bestpoint=None # for `best' which point is the best seen so far?

	# evaluate a point (return its metric score)
	def __call__(self,point):
		(point, tuples) = self.train(point)
		return self.eval(point, tuples)

	# start training process running, return (point, list of associated (pid,labels,testset,svmfile) tuples)
	def train(self,point):
		point = tuple(point) # indexable
		if point in self.cache:
			return ( point, None )	 # don't bother training
		else:
			# interpret point as parallel to flags
			parameterized_tcmd = "%s %s" % (self.tcmd, ' '.join(map( lambda (flag,value) : "%s %g" % (flag,value), gpairs(self.flags,point))))
			# pdata:  list of (pid, labels, testset, svmfile) 
			tuples = []	
			for fold in range(self.folds):
				division = CrossValidationDivision(self.dataset, self.folds, fold)	# generates separate train/test folds
				trainset,testset = list(division.trainset()),list(division.testset())
				labels  = map(lambda (x,y) : y, testset)
				testset = map(lambda (x,y) : x, testset) 
				svmfile = name_svmfile(trainset, parameterized_tcmd)
				pid = train(trainset, parameterized_tcmd, svmfile)
				tuples += [ (pid, labels, testset, svmfile) ]
			return ( point, tuples )

	# evaluate a point by its associated tuples
	def eval(self, point, tuples):
		point = tuple(point) # indexable
		if point not in self.cache:
			validationset = []
			for (pid, labels, testset, svmfile) in tuples: 
				predictions = list(interpret(predict( testset, self.pcmd, svmfile ), self.ttype))
				assert len(predictions)==len(testset)
				validationset += [gpairs(labels, gmap(lambda (y,c):y, predictions))]
				# Note: validationset  will be a list of sets (of pairs), which
				# is what the objective function expects to compute on.
				rm(svmfile) # Can't keep all of these.  No way.
			self.cache[point] = self.objective( validationset )

		return self.cache[point]

	# return best point 
	# Assume all are cached
	def best(self, point):
		point=tuple(point) # indexable
		score = self.cache[point]
		new_best = (self.bestpoint is None or self.cmp(self.cache[point],self.cache[self.bestpoint]) < 0) # A new best point?
		if point not in self.printcache:
			self.printcache[point] = gridout.write(\
				"# Grid search (%i)\t%s=%8.4f\t%s\t%s\n" % (\
				len(self.printcache)+1,\
				self.objective, self.cache[point],\
				"(*)" if new_best else "   ",\
				', '.join(map(lambda (flag,value) : "%s=%7.3g"%(flag,value), gpairs(self.flags,point)))))
		self.bestpoint = point if new_best else self.bestpoint
		return list(self.bestpoint)

# generate combinations of elements
# Example:  combinations( [  [0,1], [a,b,c] ] ) -> [0,a], [0,b], [0,c], [1,a], [1,b], [1,c], <StopIteration>
def combinations( values ):
	if values:
		for point in gridsearch_points( values[:-1] ):
			for value in values[-1]:
				yield point+[value]
	else:
		yield [] # base case

# generate neighbors of the given point (includes the point itself)
# @param origin: list of integer coordinates
def eight_neighbors(origin, radius=1):
	if origin:
		for proj in eight_neighbors(origin[1:], radius):
			yield [origin[0]  ]+proj
			for delta in range(1,radius+1):
				yield [origin[0]-delta]+proj
				yield [origin[0]+delta]+proj
	else:
		yield []

# generate neighbors of the given point (includes the point itself)
# @param origin: list of integer coordinates
def four_neighbors(origin):
	yield origin
	for d in range(len(origin)):	# dimensions
		yield origin[:d] + [origin[d] - 1] + origin[(d+1):]
		yield origin[:d] + [origin[d] + 1] + origin[(d+1):]

# small class for generating train/test folds from a data set
class CrossValidationDivision:
	def __init__(self, data, folds, fold):
		self.data = list(data)
		self.folds,self.fold = folds,fold
	def trainset(self):
		for i in range(len(self.data)):
			if i % self.folds != self.fold:
				yield self.data[i]
	def testset(self):
		for i in range(len(self.data)):
			if i % self.folds == self.fold:
				yield self.data[i]

# A fine objective function for nominal feature types:  How often correct?
class Accuracy:
	def __str__(self):
		return "Accuracy"
	def __call__(self,sets):
		c = 0 # correct
		t = 0 # total
		for set in sets:
			for (y,g) in set:
				t += 1
				c += 1 if (y==g) else 0
		return float(c) / float(t)

# Mean Squared Error:  A fine objective function for continuous features
class MSE:
	def __str__(self):
		return "MSE"
	def __call__(self,sets):
		sse = 0.0 # sum squared error
		n = 0 # total
		for set in sets:
			for (y,g) in set:
				y,g = float(y),float(g)
				sse += (y-g)*(y-g)
				n += 1
		return sse / float(n)
