#!/usr/bin/env pyth/n
# -*- coding: utf8 -*-
'''

Utilities for using a supervised classification method implemented by WEKA (e.g. decision/regression tree).
There are 

Copyright © 2010 Keith Noto.  Disributed under Gnu General Public License (see http://www.gnu.org/licenses/gpl.html).

'''
__author__ = "Keith Noto <noto@cs.tufts.edu>";
__license__ = "GPL";
__copyright__ = "Copyright © 2010 Keith Noto";
VERSION = __revision__ = "1.0"; 
# 1.0	2010/Sep	original

import sys,os,string,StringIO,time,hashlib,subprocess,tempfile,re
import parrot,lsf # may or may not be used
import storage
from util import *;

verbose = nullstream = open(os.devnull, 'w')
jar = "weka.jar" # location of .jar file for launching processes.  To be updated by user with path 

class ARFF:

	def __init__(self, (ftypes, ttype)):

		out = StringIO.StringIO()

		# create ARFF header  to be used over and over
		out.write("%% ARFF Header Created %s\n" % (time.strftime("%Y %b %d %H:%M:%S")))
		out.write("@relation %s\n" % (hash(str(ftypes)+str(ttype))))
		for (name, ftype) in map(lambda i : ("pf%i"%(i), ftypes[i]), range(len(ftypes))) + [ ("target", ttype) ]:
			if type(ftype) is list: # nominal
				out.write("@attribute %s {%s}\n" % (name, ','.join(map(str,ftype))))
			else: # continuous
				out.write("@attribute %s real\n" % (name))
		out.write("@data\n");

		self.header = out.getvalue()
		out.close()

		self.dummy = ttype[0] if type(ttype) is list else 0.0 # suggest a "dummy" placeholder target value for test sets without known targets

	def write(self, xypairs, out):
		out.write("%% ARFF File Created %s\n" % (time.strftime("%Y %b %d %H:%M:%S")))
		out.write(self.header)
		for row in map(lambda (x,y) : x + [y], xypairs):
			out.write("%s\n" % (','.join(map(lambda val : '?' if val is None else str(val), row))))

	def write_file(self, xypairs, filename ):
		out = open(filename,'w')
		self.write(xypairs, out)
		out.close()


# generic WEKA Algorithm predictor, i.e. a function that takes a test set and
# generates predictions
class Predictor:

	# create a predictor
	# command:  WEKA algorithm and any options
	# model:    learned model file
	# clean:    remove files when done?
	# pid:      process ID of training process
	def __init__(self, ttype, command, model, arff, clean, pid):
		self.ttype,self.command,self.model,self.arff,self.clean,self.pid = ttype,command,model,arff,clean,pid
	
	# Generate predictions for test set Q
	def __call__(self, Q):

		if self.pid:
			verbose.write("# Predictor (%s, %s) must wait for training process PID=%s to finish.\n" % (str(self), str(self.model), self.pid))
			os.waitpid(self.pid,0)	# wait for training process to end
			verbose.write("# Done waiting for PID=%s.\n" % (self.pid))
			self.pid = None

		# write test file
		testfile = tempfile.mktemp(".tmp.arff","weka.Q.",chkdir(storage.tmp))
		self.arff.write_file( map(lambda x:(x,self.arff.dummy),Q), filename=testfile )
	
		# run prediction process
		predictions = list(interpret(predict( self.command, testfile, self.model ), self.ttype ))

		if len(predictions) != len(Q):
			raise Exception("\n\nNumber of WEKA predictions %I does not match number of test set examples %i." % (len(predictions),len(Q)))
		
		# all done; clean up
		os.unlink(testfile)

		return predictions
			
	
	def __del__(self):
		if self.clean:
			os.unlink(self.model)

class Learner:

	# create a learner 
	def __init__(self, tcmd, pcmd, clean=True):
		self.tcmd,self.pcmd = tcmd, pcmd
		self.clean =  clean
	
	def __call__(self, (ftypes,ttype), T):

		# Remove missing target values
		T = filter(lambda (x,y) : y is not None, T)

		if not T:
			raise(Exception("\n\nCall to WEKA Learner with zero labeled training examples"))
			# Note:  One could accept this scenerio as one that probably
			# doesn't affect FRaC output by returning
			# parrot.Predictor(ttype[0]), which arbitrarily always predicts the
			# first label / minimum value.

		trainfile = tempfile.mktemp('.tmp.arff', 'weka.A.', chkdir(storage.tmp))
		modelfile = "%s/%s.%s.%s" % ( chkdir(storage.cache), hash(T), ''.join(self.tcmd.split()[1:]), self.tcmd.split()[0].split('.')[-1], ) 
	
		# Choose a name for the model file that will be the same for the exact same trainset and train command
		modeltype = self.tcmd.split()[0].split('.')[-1]
		modelcode = hash(str(T) + str(''.join(self.tcmd.split())))
		modelfile = "%s/%s.%s" % ( chkdir(storage.cache), modelcode, modeltype )

		# Create ARFF file formatter
		arff = ARFF((ftypes,ttype))

		# write training set to file
		arff.write_file( T, trainfile) 
	
		# learn model
		pid = train(self.tcmd, trainfile, modelfile)

		# clean up
		os.unlink(trainfile)
		
		# create predictor
		return Predictor(ttype, self.pcmd, modelfile, arff, self.clean, pid)


# run WEKA training process:  produce modelfile
def train(command, trainfile, modelfile):

	fork = False 
	pid = None
		
	if os.path.exists(modelfile) and os.path.getsize(modelfile):
		verbose.write("# Use existing model, %s\n" % (modelfile))

	else:

		pid = os.fork() if fork else None   # available on Unix 

		if not pid:

			if not os.path.exists(jar):
				raise Exception("\n\nRequired file %s does not exist" % (jar))

			if not os.path.exists(trainfile):
				raise Exception("\n\nTraining set file %s does not exist" % (trainfile))

			#cmd = "java -classpath %s %s -no-cv -t \"%s\" -d \"%s\"" % (jar,command,trainfile,modelfile)

			Xmx = 100 * 1024 * (os.path.getsize(trainfile) / 1024) # 100x train size for whole process
			cmd = "java -Xmx%d -classpath %s %s -no-cv -t \"%s\" -d \"%s\"" % (Xmx, jar,command,trainfile,modelfile)

			# run training process
			verbose.write("# $ %s\n" % (cmd))
			p = subprocess.Popen([cmd], shell=True, stdin=None, stdout=verbose, stderr=None, close_fds=True) 
			p.wait() 

			# to run via LSF (should be run with fork, because lsf.submit will block until finished)
			# lsf.submit(cmd, jobname="%s.%s" % (algo,os.path.basename(trainfile), jobdir=chkdir(storage.tmp)))
			
			if not os.path.exists(modelfile):

				# I don't get this:  To see the standard error produced by the
				# java command above, I have to set the stdout argument to
				# either sys.stdout or None, but then it will still write to
				# standard error.  If I set stdout=sys.stderr, it won't write
				# to it.

				verbose.flush()
				raise Exception("\n\nWEKA process failed:\n\n%s\n\n" % (cmd))
			if fork:
				sys.exit(0)  # if fork and pid, then I must be a child process (which is now done)

	return pid

# run WEKA prediction process, generate output as lines of text
def predict(command, testfile, modelfile):

	if not os.path.exists(jar):
		raise Exception("\n\nRequired file %s does not exist" % (jar))
	if not os.path.exists(testfile):
		raise Exception("\n\nweka.predict missing test set file %s\n" % (testfile))
	if not os.path.exists(modelfile):
		raise Exception("\n\nweka.predict missing model file %s\n" % (modelfile))
	
	cmd = "java -classpath %s  %s  -T \"%s\" -l \"%s\" -distribution -p 0" % (jar, command, testfile, modelfile)
	verbose.write("# $ %s\n" % (cmd))
	p = subprocess.Popen([cmd], shell=True, stdin=None, stdout=subprocess.PIPE, stderr=None, close_fds=True)

	queue = None # next line of output to generate
	for line in p.stdout:
		if queue:
			yield queue
		queue = line
	p.stdout.close()  # yield predictions w/ one-step lookahead so I can close the file before the last prediction is generated (in case calling function never asks for more)
	yield queue

# Standard interpretation of output from a WEKA prediction process.  
# WEKA algorithms with the "-distribution -p 0" options we given in `predict'
# produce output of the form:
#	...
#	 9   1:lt36.8   5:ge40.4   +   0.002,0.004,0.007,0.136,*0.851 
#	10   1:lt36.8   1:lt36.8       *0.412,0.323,0.175,0.087,0.003 
#	...
# Example number, observation, prediction (+ for error), comma-separated list of probabilities with an '*' next to the highest one.
# (Of course, regression models don't print a distribution--just an error distance)
#
# This function parses the output and produces ( prediction, distribution ) pairs
def interpret(output, ttype):

	# translate lines of output to the predictions, which are the third
	# thing in lines that begin with integers and are of the form 2:bar
	# (e.g. target value #2, which is 'bar') or else real numbers.

	translate = lambda symbol : ttype[ int(symbol.split(':')[0])-1 ] if type(ttype) is list else float(symbol)
	# above:  if the ttype is list (nominal target), find the integer index
	# (before the ':') and interpret as 1-origin index (the string value is
	# given, too, after the colon, but is sometimes truncated).  Otherwise,
	# the target is continuous, so convert to real. 

	for line in output:
		line = line.strip() # there is white space before the example numbers
		if line and '0' <= line[0] <= '9': # if line starts with example number...
			record = line.split()
			if type(ttype) is list:	 # nominal feature
				distribution = map(float,record[-1].replace('*','').split(','))
				prediction = ttype[ distribution.index(max(distribution)) ]
			else: # continuous
				prediction = float(record[2])
				distribution = None # regression model
			yield (prediction, distribution)
		# (if line doesn't begin with a number it's header information or
		# something and should be discarded)

# Codify whatever this is
def hash(thingy):
	return hashlib.md5(str(thingy)).hexdigest()

