#!/usr/bin/env python
# -*- coding: utf8 -*-
'''

Python front-end for running LOF processes.

Copyright © 2010 Keith Noto.  Disributed under Gnu General Public License (see http://www.gnu.org/licenses/gpl.html).

'''
__author__ = "Keith Noto <noto@cs.tufts.edu>"
__license__ = "GPL"
__copyright__ = "Copyright © 2010 Keith Noto"
VERSION = __revision__ = "1.0"

import sys,tempfile,string,subprocess,os,random
tmp = tempfile.gettempdir() # default temporary directory
verbose = sys.stderr # open(os.devnull,'w')

# Run LOF 
#
# trainset:  set of vectors
# testset:  If given, run semi-supervised LOF 
# algo:  one of 'lof', 'cof', 'lof.l1norm', or whatever executables are available.
def lof( trainset, testset=[], algo="lof", kLB=10, kUB=100 ):

	formatxi = str; # how to convert feature values to strings?

	D = len( trainset[0] )

	(fd, filename) = tempfile.mkstemp( suffix='.tmp', prefix='lof.', dir=tmp )
	out = os.fdopen(fd,'w')
	for x in trainset:
		out.write("%s\n" % (','.join(map(formatxi,x))))
	for x in testset:
		out.write("%s\n" % (','.join(map(formatxi,x))))
	out.close()

	cmd = "%s %d %d %d %d %s" % (algo, kLB, kUB, D, len(trainset), "%d" % (len(testset)) if testset else "") # a 5th argument (even '0') indicates a test set is present

	input = open(filename,'r')
	p = subprocess.Popen( [cmd], stdin=input, stdout=subprocess.PIPE, stderr=verbose, shell=True, close_fds=True )
	input.close()
	os.unlink( filename ) # remove  input file, no longer needed

	queue = None
	for line in p.stdout:
		if queue is not None:
			yield queue
		queue = float(line)
	p.stdout.close() # close output, end process
	yield queue

# Run LOF with feature bagging
# trials to do:  if list of integers (subset of range(trials)) is given, execute only these trials.  
#                This is useful for running different trials on different processors
def bag( trials, trainset, testset=[], algo="lof",  kLB=10, kUB=100, prngseed=92353, trials_to_do = None ):

	trainset=list(trainset)
	testset = list(testset)
	N = len(trainset[0])

	R = random.Random()
	R.seed( prngseed )
	
	sums = [ 0.0 ] * len(testset) # one running total for each testset example

	for trial in range(trials):

		F = range(N) # start with all features
		R.shuffle(F) # mix 'em up
		f = R.randint(N/2, N-1) # select this many features
		F = F[:f]  # take the first f of the mixed-up set

		if trials_to_do and trial not in trials_to_do:	
			#Hack:  spend time mixing up feature sets we're not going to do to guarantee that the seeded random.Random 
			# produces the same subsets for each trial.
			continue

		verbose.write("# Feature bagging trial %i of %i, select features %s\n" % (trial+1, trials, str(map(lambda i:i+1,F))))

		train_bag = map(lambda x : map(lambda i : x[i], F), trainset)
		test_bag = map(lambda x : map(lambda i : x[i], F), testset)
		
		i=0; 
		for score in lof(train_bag, test_bag, algo, kLB, kUB):
			sums[i] += score
			i+=1

	return sums

