/*

	2010/Feb, noto@cs.tufts.edu

	Jian Tang et al.
	"Enhancing Effectiveness of Outlier Detections for Low Density Patterns"
	PAKDD 2002.

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "oflib.h"


// distance(i,j) stored in `stored_dist' (in main).  Only allocated stored_dist where i > j 
#define dist(i,j) ((i > j) ? stored_dist[i][j] : stored_dist[j][i])

extern const char* distance_function_name();
double ac_dist(const double *c, const int n);

int main(const int argc, const char **argv) {

	FILE *verbose = stderr; 	// set to NULL to turn off verbose progress messages.

	char SYNOPSIS[2048];
	sprintf(SYNOPSIS, "Compute COF score for a set of examples using the distance function:  %s.  Input format and parameters are the same as for LOF (see lof, lof.c)", distance_function_name());
	char USAGE[2048];
	sprintf(USAGE, "Usage:  %s <k.lower.bound> <k.upper.bound> <No.Features> <No.training instances> [<No.testing.instances>]  <  <comma-separated-file>", argv[0]);

	const parameter_set param = parse_command_line(argc,argv,SYNOPSIS,USAGE);
	
	const int T = param.T; const int C0 = param.C0; const int C = param.C; 
	const int M = param.M; const int N = param.N; 
	const int kLB = param.kLB; const int kUB = param.kUB;

	const int vperiod = T/20 + 100 - (T/20)%100;	// adjust to a percentage of training set, round to nearest 100

	double **stored_dist = compute_distance(&param, verbose); // subroutine to read data, compute distance.

	int i,j,p,q,k;
	time_t t0;

	// COF
   
	if (verbose) { fprintf(verbose, "# Find %i neighbors for each of %i examples", kUB, M); }
	t0 = time(NULL);

	// find everybody's k (UB) nearest neighbors, build their trails, compute the ordered COST of those trails
	double **cost = (double**)malloc(sizeof(double*) * M);	// freed after max_COF is calculated
	int **path = (int**)malloc(sizeof(double*) * M);
	
	for (i=0; i<M; i++) {
		
		if (verbose && !(i%vperiod)) { fprintf(verbose, "."); }
	
		int *neighbors = (int*)malloc(sizeof(int) * kUB);	// freed at the bottom of the loop
		
		int met = 0; 	// neighbors found so far
		int adr = -1;	// index (in neighbors) of most distant neighbor

		for (j=0; j<N; j++) { 

			if (j==i) { continue; }

			if (met < kUB) { 
				if (met==0 || dist(i,j) >= dist(i, neighbors[adr])) { 
					adr = met;
				}
				neighbors[met++] = j; 
			} else if (dist(i,j) < dist(i,neighbors[adr])) { 
				// replace neighbor at adr with this guy (sorry, neighbor at adr, you're not a neighbor anymore)
				neighbors[ adr ] = j;

				// find new farthest neighbor
				adr = 0;
				for (k=1; k<met; k++) { 
					if (dist(i, neighbors[k]) > dist(i, neighbors[adr])) { 
						adr = k; 
					}
				}
			}

		}
		
		// neighbors now contains the k (UB) nearest neighbors, ties broken by smallest index

		// compute the trail/cost 
		path[i] = (int*)malloc(sizeof(double) * kUB);	//freed when max_COF is calculated
		cost[i] = (double*)malloc(sizeof(double) * kUB);

		int *bitmap = (int*)malloc(sizeof(int)*kUB); // bitmap[ * ] iff * is in the path, freed after loop
		for (k=0; k<kUB; k++) { bitmap[k]=0; }

		for (p=0; p<kUB; p++) {

			// find next nearest neighbor, path[p]
			cost[i][p] = 1e999;
			int best_k = -1;

			for (k=0; k<kUB; k++) { 
				// consider adding neighbors[k] to the path

				if (!bitmap[k]) { // not already in the path, 
					// compute distance from k to path

					double dist_k_path = dist(i, neighbors[k]);

					for (q=0; q<p; q++) { 
						double d = dist(path[i][q], neighbors[k]);
						dist_k_path = d < dist_k_path ? d : dist_k_path;
					}

					if (dist_k_path < cost[i][p]) {	
						// so this is the best distance to the path
						cost[i][p] = dist_k_path;
						path[i][p] = neighbors[k];
						best_k = k;
					}
				}
			}
			bitmap[ best_k ] = 1;

		}

		// now the trail (path,cost) is created.
		free(bitmap);
		free(neighbors);

	} // next trail

	for (i=0; i<M; i++) { free(stored_dist[i]); } free(stored_dist);
	
	if (verbose) { fprintf(verbose, "%.0lf second(s).\n# Compute COF for %i <= k <= %i", difftime(time(NULL),t0), kLB, kUB); }
	t0 = time(NULL);

	double *max_COF = (double*)malloc(sizeof(double) * C);	// freed at end of main
	for (i=0; i<C; i++) { max_COF[i] = -1e999; }

	for (k=kLB; k<=kUB; k++) { 
    
		if (verbose) { fprintf(verbose, "."); }
		
		// compute the ac-dist for each example
		double *ac_dists = (double*)malloc(sizeof(double) * M);	// freed at bottom of loop
		for (i=0; i<M; i++) { 
			ac_dists[i] = ac_dist(cost[i], k+1); 
		}

		// for each test example,  compute ac-dist and average
		//	ac-dist for elements in its path

		for (i=C0; i<C0+C; i++) { 
		
			double sum_ac_dist_o = 0.0;
			for (p=0; p<k; p++) { 
				sum_ac_dist_o += ac_dists[ path[i][p] ];
			}
			
			double COF = k * ac_dists[ i ] 
			    / //  ---------------------
			             sum_ac_dist_o;

			// keep the largest COF
			max_COF[i-C0] = COF > max_COF[i-C0] ? COF : max_COF[i-C0];
		}
	
		free(ac_dists);

	}
	
	if (verbose) { fprintf(verbose, "%.0lf second(s).\n", difftime(time(NULL),t0)); }

	for (i=0; i<M; i++) { 
		free(path[i]);
		free(cost[i]);
	}
	free(path);
	free(cost);

	// print the (maximum) COF for each test instance
	for (i=0; i<C; i++) { 
		fprintf(stdout, "%10.10g\n", max_COF[i]); 
	}

	free(max_COF);

	return 0;

} // main 

double ac_dist(const double *dist_e, const int r) {
	int i;
	double sum = 0.0;
	for (i=1; i<r; i++) { 
		sum += (r - i) * dist_e[i-1];
	}
	return sum * 2 / (r * (r-1));
}	

