#!/usr/bin/env python
# -*- coding: utf8 -*-
'''

Aggregate functions.

Copyright © 2009 Keith Noto.  Disributed under Gnu General Public License (see http://www.gnu.org/licenses/gpl.html).

'''
__author__ = "Keith Noto <knoto@cs.ucsd.edu>";
__license__ = "GPL";
#__copyright__ = "Copyright (C) 2009, Keith Noto";
__copyright__ = "Copyright © 2009 Keith Noto";
VERSION = __revision__ = "1.02";

# 1.0	original
# 1.01	fixed bug in 'argmax' (it was more like just 'max'), 2009/July/28
# 1.02	added Bessel's correction option to variance and std. dev. calculations. 2009/Nov/06

import sys, string, math;

class Undefined:
	def __str__(self):
		return "undefined";

undefined = Undefined();	# THE undefined distinguished value

# @see passes
class IterCounter:
	def __init__(self):
		self.calls=0;
	def __iter__(self):
		self.calls+=1;
		return self;
	def next(self):
		raise StopIteration;

# count how many times aggregate function iterates through a set of data.  This
# is useful if you're trying to decide whether to read data into memory or not.
# Be aware that `function' is passed an "empty" iteratable and may raise an exception.
def passes(function):
	ic = IterCounter();
	try:
		function(ic);
	except:
		pass;
	return ic.calls;
	
def sum(X):
	return reduce(lambda a,x : a+x, X, 0.0);

def product(X):
	return reduce(lambda a,x: a*x, X, 1.0);

def arithmetic_average(X):
	( sum, N ) = reduce(lambda (a,N),x : (a+x,N+1) , X, (0.0,0)); 
	return undefined if not N else sum/N;
mean = avg = arithmetic_average;	# alternate name

def multiplicative_average(X):
	( product, N ) = reduce( lambda (a,N),x : (a*x,N+1), X, (1.0,0));
	return undefined if not N else math.pow(product , 1.0/N );

def harmonic_average(X):
	( sum, N ) = reduce( lambda (a,N),x : (a+(1.0/x),N+1), X, (0.0,0));
	return undefined if not N else N/sum;

# return list of x in X s.t. f(x) is greatest
def argmax(X, f, cmp = lambda a,b : a-b):
	maxx,maxf = [],None;
	for x in X:
		fx = f(x);
		if maxf is None:
			maxx,maxf = [x],fx;
		else:
			c = cmp(fx,maxf);
			if c > 0:
				maxx,maxf = [x],fx;
			elif c == 0:
				maxx = maxx+[x]
	return maxx;
			
# which zero-origin index(s) is most?
def imax(X,cmp=lambda a,b : a-b):
	maxi,maxx = [],None;
	index = 0;
	for x in X:
		if maxx is None:
			maxi,maxx=[index],x;
		else:
			c = cmp(x,maxx);
			if c > 0:
				maxi,maxx = [index],x;
			elif c==0:
				maxi,maxx = maxi+[index],x;
		index+=1;
	return maxi;	

def count(X):
	return reduce(lambda a,x:a+1, X, 0 );

# utility
def ssd_n(X,mu=None):
	# return (sum of squared difference from the mean, N)
	mu = arithmetic_average(X) if mu is None else mu;	# pass #1
	return reduce(lambda (a,N),x : (a+((mu-x)*(mu-x)) , N+1), X, (0.0,0)); 

def variance(X, bessels_correction, mu=None):
	# mean squared difference
	( ssd, N ) = ssd_n(X,mu);
	return undefined if not N else ssd / ((N-1) if bessels_correction else N);

# @param bessels_correction Use N-1 in the formula (sample std. dev.) instead of N (population std. dev)
def stdev(X, bessels_correction, mu=None):
	# mean squared difference
	( ssd, N ) = ssd_n(X,mu);
	return undefined if not N else math.sqrt( ssd / ((N-1) if bessels_correction else N) );

def median(X):
	M=[];
	for x in X:
		M+=[x];
	M.sort(); 
	if M:
		return M[ len(M)/2 ];
	else:
		return undefined;

def mode(X):
	C = {};	# C[x]->count
	mode = C_mode = None;
	for x in X:
		C_x = C[x] = C.get(x,0) + 1;
		if (mode is None or C_x > C_mode):
			mode = x;
			C_mode = C_x;
	return mode;


