
import os,re,sys,string
import metric,Timer
from util import *

verbose = sys.stderr

# Columns are:  ( (files,...), Title, Is-baseline?, Is-normalscore? )
# where files is a list of files containing anomaly detection scores (multiple files are pasted and added together), 
# title is a header entry, 
# Is-baseline means it's okay for there to be fewer predictions than labels (some of those extra testset anomalies are used for training), 
# and Is-normalscore means the scores should be interpreted in reverse (lower => more likely anomalous)

COLUMNS = [ \
(['baseline.linearsvm'],'Baseline (linear kernel SVM)', True, False), \
(['baseline.rbfsvm'],   'Baseline (RBF kernel SVM)', True, False), \
(['baseline.tree'],     'Baseline (C4.5 Decision Tree)', True, False), \
\
(['one-class-svm.linear'], 'One-Class SVM (linear kernel)', False, True), \
(['one-class-svm.rbf'],    'One-Class SVM (RBF kernel)', False, True), \
\
(['lof.10-100'],         'LOF', False, False), \
\
(['lof.10-100.100bags'], 'LOF w/ Feature Bagging', False, False), \
(['cof.10-100'],         'COF', False, False), \
(['cof.10-100.100bags'], 'COF w/ Feature Bagging', False, False), \
\
(['cfa2.nbc.amc'],       'Cross-Feature Analysis (Naive Bayes, Average Match Count)', False, True), \
(['cfa2.ripper.amc'],    'Cross-Feature Analysis (RIPPER, Average Match Count)', False, True), \
(['cfa2.tree.amc'],      'Cross-Feature Analysis (C4.5, Average Match Count)', False, True), \
(['cfa2.linearsvm.amc'], 'Cross-Feature Analysis (Linear kernel SVM, Average Match Count)', False, True), \
(['cfa2.rbfsvm.amc'],    'Cross-Feature Analysis (RBF kernel SVM, Average Match Count)', False, True), \
(['cfa2.tree.amc','cfa2.linearsvm.amc','cfa2.rbfsvm.amc'], 'Cross-Feature Analysis (Tree+SVM+SVM, Average Match Count)', False, True),\
\
(['cfa2.nbc.apx'],       'Cross-Feature Analysis (Naive Bayes, Average Probability)', False, True), \
(['cfa2.ripper.apx'],    'Cross-Feature Analysis (RIPPER, Average Probability)', False, True), \
(['cfa2.tree.apx'],      'Cross-Feature Analysis (C4.5, Average Probability)', False, True), \
(['cfa2.linearsvm.apx'], 'Cross-Feature Analysis (Linear kernel SVM w/ Platt Scaling, Average Probability)', False, True), \
(['cfa2.rbfsvm.apx'],    'Cross-Feature Analysis (RBF kernel SVM w/ Platt Scaling, Average Probability)', False, True), \
(['cfa2.tree.apx','cfa2.linearsvm.apx','cfa2.rbfsvm.apx'], 'Cross-Feature Analysis (Tree+SVM+SVM, Average Probability)', False, True),\
\
(['frac.linearsvm'],    'FRaC (linear kernel SVM)', False, False), \
(['frac.rbfsvm'],       'FRaC (RBF kernel SVM)', False, False), \
(['frac.tree'],         'FRaC (C4.5 or Fast Regression Tree)', False, False), \
(['frac.tree','frac.rbfsvm','frac.linearsvm'], 'FRaC (Tree+SVM+SVM)', False, False), \
\
(['frac.q5.2.nbc'],       'FRaC (discretized features, Naive Bayes)', False, False), \
(['frac.q5.2.ripper'],    'FRaC (discretized features, RIPPER)', False, False), \
(['frac.q5.2.tree'],      'FRaC (discretized features, C4.5)', False, False), \
(['frac.q5.2.linearsvm'], 'FRaC (discretized features, linear kernel SVM)', False, False), \
(['frac.q5.2.rbfsvm'],    'FRaC (discretized features, RBF kernel SVM)', False, False), \
(['frac.q5.2.tree','frac.q5.2.rbfsvm','frac.q5.2.linearsvm'], 'FRaC (discretized features, Tree+SVM+SVM)', False, False), \
\
]

def method_description( code ):
	for (files, desc, baseline, reversed) in COLUMNS:
		if code == '+'.join(files):
			return desc
	return None

def gen_auc(rpath, dataset, i, methods, baseline, normalscore):

	for (spath,seed) in files(rpath):

		yfile = "%s/%s/labels" % (spath,dataset)
		gfiles = list(map(lambda method : "%s/%s%s/%s" % (spath, dataset, ".i%d"%(i) if i else "", method), methods)) # (e.g., ecoli.i13 for i==13)

		if all(map(os.path.exists,[yfile]+gfiles)) and all(map(os.path.getsize,[yfile]+gfiles)):

			G = map(sum, piecewise(map(lambda gfile : read_numbers(gfile), gfiles)))
			Y = list(read_labels(yfile))

			if baseline:
				Y = Y[:len(G)]  # supervised baseline methods have fewer predictions
			if normalscore:
				G = map(lambda score : -score, G) # reverse sign of predictions for methods that produce higher values for normals
			if len(G) != len(Y):
				#sys.stderr.write("\n\ngfiles=%s\n\n" % (str(gfiles)))#
				#raise Exception("\n\nIncompatible labels/prediction files:\n\t%s (%d)\n\t[all prediction files] (%d): %s,... (%d files)\n\n" % (yfile,len(Y),len(G), ','.join(gfiles[:2]), len(gfiles) ))
				sys.stderr.write("#\n\nERROR:\n\nIncompatible labels/prediction files:\n\t%s (%d)\n\t[all prediction files] (%d): %s,... (%d files)\n\n" % (yfile,len(Y),len(G), ','.join(gfiles[:2]), len(gfiles) ))
			else:
				yield (seed, metric.auc(metric.roc(gpairs(Y,G))))

	
def gen_datasets(rpath):
	D = {}
	for (spath,seed) in files(rpath):
		for (dpath,dataset) in files(spath):
			if dataset not in D:
				yield dataset
			D[dataset]=True

def gen_i(rpath, dataset):
	D = { } # set of i values seen
	for (spath,seed) in files(rpath):
		for (dpath,experiment) in files(spath):
			if re.match( "%s\\.i[0-9][0-9]*"%(dataset), experiment ):
				i = int(experiment.split('.i')[1])
			else:
				i = 0
			if i not in D:
				yield i
			D[i]=True

# generate (fullpath,basename) pairs for all files in a directory
def files(directory):
	if os.path.exists(directory):
		for file in os.listdir(directory):
			yield ( "%s/%s" % (directory,file), file )

def read_file(filename):
	input = open(filename,'r')
	for line in input:
		yield line
	input.close()

def read_labels(filename):
	for line in read_file(filename):
		[ ID, label, anomaly ] = line.split()
		yield True if int(anomaly) else False
	
# numbers file may be a column of numbers, or a matrix.
# in either case, generate the aggregate over each row
def read_numbers(filename, ag=sum):
	for line in read_file(filename):
		yield ag( map(float,line.split()))

def piecewise(X):
	its = map(iter,X)
	while True:
		try:
			yield map(next,its)
		except StopIteration:
			break
	

# This is made elaborate so one can identify the member not the right length
def piecewise(X):
	its = map(iter, X) # that's a python variable name, 'it's...'
	while True:
		x = [ ]
		for i in range(len(its)):
			try:
				x += [ next(its[i]) ]
			except StopIteration:
				if i==0:
					for j in range(1,len(its)):
						try:
							next(its[j])
							raise Exception("Argument #%d to piecewise is longer than argument #1" % (j+1))
						except StopIteration:
							pass
					break
				else:
					raise Exception("Argument #%d to piecewise is shorter than argument #1" % (i+1))
		if x:
			yield x
		else:
			break

def write(output, datasets, rpath, iset=None):
	output.write("#Dataset\ti\tMethod\tSeed\tAUC\n")
	for dataset in datasets if datasets else sorted(gen_datasets(rpath)):
		t = Timer.Timer()
		verbose.write("%s"%(dataset))
		for i in iset if iset else sorted(gen_i(rpath, dataset)):
			for (files, method, baseline, reversed) in COLUMNS:
				verbose.write(".")
				code = "+".join(files)
				for (seed,AUC) in gen_auc( rpath, dataset, i, files, baseline, reversed  ):
					output.write("%s\n" % ('\t'.join(map(str, [ dataset, 0 if i is None else i, code, seed, AUC ] ))))
		verbose.write("%s\n" % (t))

# Read auc file data into memory.
# datasets: limit to these datasets
# ivalues: limit to these values of i
# return:  database[ dataset ][ i ][ method ][ seed ] -> AUC
def read(input, datasets=None, ivalues=None ):
	database = { }
	for line in input:
		line = line.split('#')[0]
		if line:
			line = line.split('\t')
			line = map(string.strip, line)
			[ dataset, i, method, seed, AUC ] = line
			i = 0 if i is None or i=='None' else int(i)
			seed = int(seed)
			AUC = float(AUC)
			if datasets and dataset not in datasets:
				continue
			if ivalues is not None and i not in ivalues:
				continue
			if dataset not in database:
				database[dataset] = { }
			if i not in database[dataset]:
				database[dataset][i] = { }
			if method not in database[dataset][i]:
				database[dataset][i][method] = { }
			if seed not in database[dataset][i][method]:
				database[ dataset ][ i ][ method ][ seed ] = { }
			assert seed not in database[ dataset ][ i ][ method ][ seed ] or database[ dataset ][ i ][ method ][ seed ] == AUC # Ignore double entry unless it has a different AUC
			database[ dataset ][ i ][ method ][ seed ] = AUC
	return database

