#!/usr/bin/env python
# -*- coding: utf8 -*-
'''
Scoring functions.  

Usually, the domain is a set of (observed, predicted) pairs.

There are regression metrics (i.e. treat observed,predicted as reals),
accuracy metrics (i.e. treat observed,predicted as exact values), and 
binary metrics (i.e. treat observed,predicted as either positive or negative).

For binary metrics, it is a good idea to map pairs to (True=positive,
False=negative) or something interpretable as binary labels (The `ispos'
function does this interpretation).

Copyright © 2009 Keith Noto.  Disributed under Gnu General Public License (see http://www.gnu.org/licenses/gpl.html).

'''
__author__ = "Keith Noto <noto@cs.tufts.edu>";
__license__ = "GPL";
__copyright__ = "Copyright © 2009 Keith Noto";
VERSION = __revision__ = "1.01";
# 1.01	2010/Jan/13	fixed bug in group sort, used by ROC and PR curves.  Edited auc to return None iff area is undefined

import ag,confusion;

# sum squared error (a regression aggregate statistic)
def sse(pairs):
	sse = 0.0;
	for (obs,pred) in pairs:
		obs = float(obs);
		pred = float(pred);
		sse += (obs-pred)*(obs-pred);
	return sse;

# mean squared error (a regression aggregate statistic)
def mse(pairs):
	sse = 0.0;
	N = 0;
	for (obs,pred) in pairs:
		obs = float(obs);
		pred = float(pred);
		sse += (obs-pred)*(obs-pred);
		N += 1;
	return sse/float(N);


# Pearson corrlation
def correlation(pairs):

	X = []; 
	Y = []; 

	for (x,y) in pairs:
		X += [ float(x) ];
		Y += [ float(y) ]; 

	Xbar = ag.mean(X);
	Ybar = ag.mean(Y);

	if (not Xbar or not Ybar):
		return ag.undefined;
	else:
		s_X = ag.stdev(X, Xbar);
		s_Y = ag.stdev(Y, Ybar);
		if (s_X==0.0 or s_Y==0.0):
			return ag.undefined;
		else:
			sigma_term = 0.0;
			for i in range(0, len(X)):
				sigma_term += ((X[i] - Xbar)/s_X) * ((Y[i] - Ybar)/s_Y); 
			r = (1.0/len(X)) * sigma_term; 
			return r;


# Default function to determine if a label/target should be considered positive
# (for binary functions, e.g. for precision/recall)
def ispos(target):
	if target is True:
		return True;
	elif target is False:
		return False;
	try:
		# numerical values are positive if they're positive
		return float(target) > 0.0;
	except:
		return ((str(target))[0]) in "+pP";

# Generate binary version of pairs
def binary(pairs):
	for (o,p) in pairs:
		yield (ispos(o),ispos(p)); 

# compute accuracy as |correct| / |examples|
def accuracy(pairs):
	return confusion.accuracy(confusion.confusion(pairs)); 

# compute precision as |correctly predicted| / |predicted|
def precision(pairs):
	return confusion.precision(confusion.confusion(binary(pairs))); 

# compute recall as |correctly predicted| / |positive|
def recall(pairs):
	return confusion.recall(confusion.confusion(binary(pairs))); 

# compute f-measure over pairs
def f_measure(pairs,beta):
	return confusion.f_measure(confusion.confusion(binary(pairs)),beta); 

# compute f1 
def f1(pairs):
	return f_measure(pairs,1.0);

# In-place sort (obs,pred) pairs by decreasing pred @return argument, which is sorted
def csort(pairs):
	pairs.sort( lambda (o1,p1),(o2,p2) : -1 if p1>p2 else 1 if p1<p2 else 0 );
	return pairs;




# Generate the set of ordered (x,y)=(FPR,TPR) pairs that make up the ROC curve.
# 
# Important "gotcha:"  If all predictions are the same (e.g. the hypothesis is
# as simple as possible, "everything's negative"), then the sorted order of
# predictions is arbitrary.  For this reason, ROCs are generated by groups of 
# observed labels assigned to the same prediction value, and those groups are sorted
# in descreasing order of prediction confidence.
#
# @param pairs Set of (observed label,predicted label)
# @linear_interpolate Plot only points after groups that contain some positive examples
def roc(pairs, linear_interpolate=False):
	groups = group_sort_pairs(pairs)
	if linear_interpolate:
		groups = positive_runs(groups)
	return roc_label_groups(groups)

# Generate the set of (x,y)=(FPR,TPR) pairs that make up the Precision/Recall curve.
# @param pairs Set of (observed label,predicted label)
# @param sawtooth Allow the "sawtooth" effect by plotting all points (alternative is to have curve decrease monotonically)
# @linear_interpolate Plot only points after groups that contain some positive examples
# @see roc 
def prc(pairs, sawtooth=False, linear_interpolate=False):
	groups = group_sort_pairs(pairs)
	if linear_interpolate:
		groups = list(positive_runs(groups))
	groups.reverse();	# go in reverse order so I can delete the sawtooth
	return prc_label_groups((groups), sawtooth);
rpc=prc; # same name

# Generate the set of ordered (x,y)=(FPR,TPR) pairs that make up the ROC curve 
# from a list of observed Boolean labels in the order of highest confidence to lowest.
# @see roc
def roc_label_groups(sorted_label_groups):
	sorted_label_groups = list(sorted_label_groups)

	#count labels
	(P,N)=(0,0);
	for group in sorted_label_groups:
		for x in group:
			(P,N) = (P+1,N) if x else (P,N+1);
				
	(P,N) = (float(P), float(N)); # no need to do conversions later
	TP = FP = 0;
	
	yield (0,0); 
	for group in sorted_label_groups:
		for label in group:
			if label:
				TP = TP + 1.0;
			else:
				FP = FP + 1.0;
			x = FPR = FP/N if N else 0; 
			y = TPR = TP/P if P else 0;
		yield (x,y);

# Generate the set of (x,y)=(FPR,TPR) pairs that make up the Precision/Recall
# curve from a list of observed labels in the order of lowest confidence to
# highest confidence.
# @see prc 
# @see roc
# @param sawtooth Allow the "sawtooth" effect by plotting all points (alternative is to have curve decrease monotonically)'
def prc_label_groups(reverse_sorted_label_groups, sawtooth=False):
	reverse_sorted_label_groups=list(reverse_sorted_label_groups)

	# count labels
	(P,N)=(0,0);
	for group in reverse_sorted_label_groups:
		for x in group:
			(P,N) = (P+1,N) if x else (P,N+1);
	
	# map TP, FP to (x=recall, y=precision)
	pointmap = lambda TP,FP : (TP / float(P) if P else 0 , float(TP) / float(TP+FP) if (TP+FP > 0) else 1.0 ); 

	(_,best_precision) = pointmap(P,N);
	TP = P; 
	FP = N;

	yield pointmap(TP,FP); # the ``everything'' hypothesis
	for group in reverse_sorted_label_groups:
		for label in group:
			if label:	# if label is positive
				TP -= 1;	# lose a true positive
			else:
				FP -= 1;
			(recall,precision) = pointmap(TP,FP); 
			best_precision = max(precision,best_precision);
			(x,y) = (recall, precision if sawtooth else best_precision);
		yield (x,y);

# Subroutine for roc and prc:  Map a set of (obs,pred) pairs to  a set of 
# Boolean(obs) grouped by pred, sorted by pred (descending)
def group_sort_pairs(pairs):
	pairs = list(pairs);	# read into memory
	prev_pred = None;
	group = [];
	groups = []; # sorted label groups
	for (obs,pred) in map(lambda (o,p) : (ispos(o),p), csort(pairs)):
		if group and pred!=prev_pred:
			groups += [ group ];
			group = [ obs ];	# 2010/Jan/13 forgot to include obs in this group
		else:
			group += [obs];
		prev_pred = pred;
	if group:
		groups += [group];
	return groups;

# Combine groups so that all groups include some positives (i.e. do not plot
# points that only reduce
# accuaracy
def positive_runs(groups):
	n = []
	for group in groups:
		if any(group):
			yield group + n;
			n = []
		else:
			n += group;
	if n:
		yield n;
		


# compute the area under the curve
def auc(points):
	area = None; # undefined (2010/Jan/13)
	p1 = None;	# previous point
	for (x2,y2) in points:
		if (p1 is not None):
			(x1,y1) = p1;
			area = 0.0 if area is None else area;	# reset area 
			area += 0.5 * abs(x2-x1) * abs(y2-y1);	# triangle
			area += abs(x2-x1) * min(y2,y1);		# rectangle 
		p1 = (x2,y2);
	return area;

# replace scores with rank in each set of pairs, then pool them all together
def pool_by_rank(pairsets):
	for pairs in pairsets:
		rank = 1
		pairs = list(pairs)
		for group in reversed(group_sort_pairs(pairs)):
			for obs in group:
				yield (obs,rank);
			rank += len(list(group))

